import mimetypes
import re
import tempfile
import time
from pathlib import Path
from typing import Any, Literal
from urllib.parse import urlparse

import requests

from isekai.types import BlobResource, Key, PathFileProxy, TextResource

# Text MIME types that should be treated as text data
TEXT_MIME_TYPES = {
    "application/json",
    "application/xml",
    "application/javascript",
}


class BaseExtractor:
    def extract(
        self, key: Key, metadata: dict[str, Any] | None = None
    ) -> TextResource | BlobResource | None:
        return None


class HTTPExtractor(BaseExtractor):
    def __init__(
        self,
        max_retries: int = 3,
        max_delay: float = 60.0,
        timeout: int = 30,
        no_retry_status_codes: set[int] | None = None,
    ):
        self.max_retries = max_retries
        self.max_delay = max_delay
        self.timeout = timeout
        self.no_retry_status_codes = no_retry_status_codes or {404}

    def extract(
        self, key: Key, metadata: dict[str, Any] | None = None
    ) -> TextResource | BlobResource | None:
        # We only handle keys of type "url"
        if key.type != "url":
            return None

        url = key.value

        response = self._make_request_with_backoff(url)

        content_type = response.headers.get("Content-Type", "application/octet-stream")
        mime_type = content_type.split(";")[0]
        data_type = self._detect_data_type(mime_type)

        # Create metadata with response headers
        metadata = {"response_headers": dict(response.headers)}

        if data_type == "text":
            return TextResource(
                mime_type=mime_type, text=response.text, metadata=metadata
            )
        else:
            filename = self._infer_filename(url, response, mime_type)
            # Create a temporary file to store the blob data
            temp_file = tempfile.NamedTemporaryFile(delete=False, suffix=f"_{filename}")
            temp_file.write(response.content)
            temp_file.close()

            return BlobResource(
                mime_type=mime_type,
                filename=filename,
                file_ref=PathFileProxy(path=Path(temp_file.name)),
                metadata=metadata,
            )

    def _make_request_with_backoff(self, url: str) -> requests.Response:
        """Make HTTP request with exponential backoff retry logic."""
        for attempt in range(self.max_retries + 1):
            try:
                response = requests.get(url, timeout=self.timeout)
                response.raise_for_status()
                return response
            except requests.exceptions.HTTPError as e:
                # Don't retry configured status codes, raise immediately
                if (
                    e.response is not None
                    and e.response.status_code in self.no_retry_status_codes
                ):
                    raise e

                if attempt < self.max_retries:
                    delay = min(2**attempt, self.max_delay)
                    time.sleep(delay)
                    continue

                # Last attempt failed, re-raise the exception
                raise e
            except requests.exceptions.RequestException as e:
                if attempt < self.max_retries:
                    delay = min(2**attempt, self.max_delay)
                    time.sleep(delay)
                    continue

                # Last attempt failed, re-raise the exception
                raise e

        # This is mostly for type checkers; we should never reach here
        raise RuntimeError("Unreachable code reached in _make_request_with_backoff")

    def _detect_data_type(self, content_type: str) -> Literal["text", "blob"]:
        # Check if it's a text MIME type
        if content_type.startswith("text/") or content_type in TEXT_MIME_TYPES:
            return "text"

        # Otherwise, treat it as binary data
        return "blob"

    def _infer_filename(
        self, url: str, response: requests.Response, mime_type: str
    ) -> str:
        """Infer filename from URL, response headers, or MIME type."""

        # 1. Try Content-Disposition header
        content_disposition = response.headers.get("Content-Disposition", "")
        if content_disposition:
            filename_match = re.search(
                r'filename[*]?=(?:"([^"]+)"|\'([^\']+)\'|([^;\s]+))',
                content_disposition,
            )
            if filename_match:
                filename = (
                    filename_match.group(1)
                    or filename_match.group(2)
                    or filename_match.group(3)
                )
                if filename:
                    return filename

        # 2. Try URL path
        parsed_url = urlparse(url)
        if parsed_url.path:
            path_parts = parsed_url.path.split("/")
            for part in reversed(path_parts):
                if part and "." in part:
                    return part

        # 3. Generate from MIME type
        extension = mimetypes.guess_extension(mime_type)
        if extension:
            # Remove the leading dot and handle some common cases
            extension = extension.lstrip(".")
            # mimetypes sometimes returns .jpe for image/jpeg, prefer .jpg
            if extension == "jpe":
                extension = "jpg"
        else:
            extension = "bin"

        # Try to use the last path segment as base filename
        parsed_url = urlparse(url)
        if parsed_url.path:
            path_parts = parsed_url.path.strip("/").split("/")
            if path_parts and path_parts[-1]:
                base_name = path_parts[-1]
                # Remove existing extension if present
                if "." in base_name:
                    base_name = base_name.rsplit(".", 1)[0]
                return f"{base_name}.{extension}"

        return f"file.{extension}"



import pytest

from isekai.utils.graphs import (
    build_condensation,
    resolve_build_order,
    tarjan_scc,
    topo_sort,
)


class TestTarjanSCC:
    def test_single_node(self):
        nodes = ["a"]
        edges = []
        sccs, node_to_scc = tarjan_scc(nodes, edges)

        assert len(sccs) == 1
        assert sccs[0] == ["a"]
        assert node_to_scc["a"] == 0

    def test_two_disconnected_nodes(self):
        nodes = ["a", "b"]
        edges = []
        sccs, node_to_scc = tarjan_scc(nodes, edges)

        assert len(sccs) == 2
        assert {frozenset(scc) for scc in sccs} == {frozenset(["a"]), frozenset(["b"])}
        assert len(set(node_to_scc.values())) == 2

    def test_simple_cycle(self):
        nodes = ["a", "b", "c"]
        edges = [("a", "b"), ("b", "c"), ("c", "a")]
        sccs, node_to_scc = tarjan_scc(nodes, edges)

        assert len(sccs) == 1
        assert set(sccs[0]) == {"a", "b", "c"}
        assert node_to_scc["a"] == node_to_scc["b"] == node_to_scc["c"] == 0

    def test_acyclic_graph(self):
        nodes = ["a", "b", "c"]
        edges = [("a", "b"), ("b", "c")]
        sccs, node_to_scc = tarjan_scc(nodes, edges)

        assert len(sccs) == 3
        assert all(len(scc) == 1 for scc in sccs)
        assert len(set(node_to_scc.values())) == 3

    def test_multiple_sccs(self):
        nodes = ["a", "b", "c", "d", "e"]
        edges = [("a", "b"), ("b", "a"), ("b", "c"), ("c", "d"), ("d", "e"), ("e", "d")]
        sccs, _ = tarjan_scc(nodes, edges)

        assert len(sccs) == 3
        scc_sets = {frozenset(scc) for scc in sccs}
        expected_sets = {frozenset(["a", "b"]), frozenset(["c"]), frozenset(["d", "e"])}
        assert scc_sets == expected_sets


class TestCondensation:
    def test_single_node(self):
        edges = []
        comp_id = {"a": 0}
        k = 1
        condensation = build_condensation(edges, comp_id, k)

        assert len(condensation) == 1
        assert 0 in condensation
        assert condensation[0] == set()

    def test_two_disconnected_nodes(self):
        edges = []
        comp_id = {"a": 0, "b": 1}
        k = 2
        condensation = build_condensation(edges, comp_id, k)

        assert len(condensation) == 2
        assert condensation[0] == set()
        assert condensation[1] == set()

    def test_simple_cycle(self):
        edges = [("a", "b"), ("b", "c"), ("c", "a")]
        comp_id = {"a": 0, "b": 0, "c": 0}
        k = 1
        condensation = build_condensation(edges, comp_id, k)

        assert len(condensation) == 1
        assert 0 in condensation
        assert condensation[0] == set()

    def test_acyclic_graph(self):
        edges = [("a", "b"), ("b", "c")]
        comp_id = {"a": 0, "b": 1, "c": 2}
        k = 3
        condensation = build_condensation(edges, comp_id, k)

        assert len(condensation) == 3
        total_edges = sum(len(neighbors) for neighbors in condensation.values())
        assert total_edges == 2

    def test_multiple_sccs_with_connections(self):
        edges = [("a", "b"), ("b", "a"), ("b", "c"), ("c", "d"), ("d", "e"), ("e", "d")]
        comp_id = {"a": 0, "b": 0, "c": 1, "d": 2, "e": 2}
        k = 3
        condensation = build_condensation(edges, comp_id, k)

        assert len(condensation) == 3
        assert 1 in condensation[0]  # SCC {a,b} -> SCC {c}
        assert 2 in condensation[1]  # SCC {c} -> SCC {d,e}
        total_edges = sum(len(neighbors) for neighbors in condensation.values())
        assert total_edges == 2


class TestBuildOrder:
    def test_simple_linear_dependency(self):
        """Test a simple chain: home_page -> img_a -> favicon"""
        nodes = ["home_page", "img_a", "favicon"]
        edges = [("home_page", "img_a"), ("img_a", "favicon")]

        sccs, comp_id = tarjan_scc(nodes, edges)
        condensation = build_condensation(edges, comp_id, len(sccs))
        build_order = topo_sort(condensation)

        # Should have 3 separate components (no cycles)
        assert len(sccs) == 3
        assert all(len(scc) == 1 for scc in sccs)

        # Build order should respect dependencies
        home_comp = comp_id["home_page"]
        img_comp = comp_id["img_a"]
        favicon_comp = comp_id["favicon"]

        home_pos = build_order.index(home_comp)
        img_pos = build_order.index(img_comp)
        favicon_pos = build_order.index(favicon_comp)

        assert favicon_pos < img_pos < home_pos

    def test_no_dependencies(self):
        """Test independent resources that can be built in any order"""
        nodes = ["home_page", "about_page", "contact_page"]
        edges = []

        sccs, comp_id = tarjan_scc(nodes, edges)
        condensation = build_condensation(edges, comp_id, len(sccs))
        build_order = topo_sort(condensation)

        # Should have 3 separate components
        assert len(sccs) == 3
        assert len(build_order) == 3
        # All can be built independently
        assert len(condensation) == 3
        assert all(len(neighbors) == 0 for neighbors in condensation.values())

    def test_simple_cycle_dependency(self):
        """Test cyclic dependency that must be built together"""
        nodes = ["page_a", "page_b", "page_c"]
        edges = [("page_a", "page_b"), ("page_b", "page_c"), ("page_c", "page_a")]

        sccs, comp_id = tarjan_scc(nodes, edges)
        condensation = build_condensation(edges, comp_id, len(sccs))
        build_order = topo_sort(condensation)

        # Should have 1 component containing all nodes
        assert len(sccs) == 1
        assert set(sccs[0]) == {"page_a", "page_b", "page_c"}
        assert len(build_order) == 1

    def test_mixed_dependencies_with_scc(self):
        """Test resources with both linear dependencies and cycles"""
        nodes = ["home_page", "img_a", "img_b", "style_a", "style_b", "footer"]
        edges = [
            ("home_page", "img_a"),  # home depends on img_a
            ("img_a", "img_b"),  # img_a depends on img_b (cycle)
            ("img_b", "img_a"),  # img_b depends on img_a (cycle)
            ("img_a", "style_a"),  # img_a depends on style_a (cycle)
            ("style_a", "style_b"),  # style_a depends on style_b (cycle)
            ("style_b", "img_a"),  # style_b depends on img_a (cycle)
            ("style_a", "footer"),  # style_a depends on footer
        ]

        sccs, comp_id = tarjan_scc(nodes, edges)
        condensation = build_condensation(edges, comp_id, len(sccs))
        build_order = topo_sort(condensation)

        # Should have 3 components: {home_page}, {img_a, img_b, style_a, style_b}, {footer}
        assert len(sccs) == 3

        # Find the large SCC
        large_scc = max(sccs, key=len)
        assert len(large_scc) == 4
        assert set(large_scc) == {"img_a", "img_b", "style_a", "style_b"}

        # Verify build order respects dependencies
        home_comp = comp_id["home_page"]
        footer_comp = comp_id["footer"]
        scc_comp = comp_id["img_a"]  # All SCC members have same comp_id

        home_pos = build_order.index(home_comp)
        footer_pos = build_order.index(footer_comp)
        scc_pos = build_order.index(scc_comp)

        # Footer must be built before the SCC, SCC before home
        assert footer_pos < scc_pos < home_pos

    def test_diamond_dependency(self):
        """Test diamond-shaped dependency graph"""
        nodes = ["root", "left", "right", "leaf"]
        edges = [
            ("root", "left"),
            ("root", "right"),
            ("left", "leaf"),
            ("right", "leaf"),
        ]

        sccs, comp_id = tarjan_scc(nodes, edges)
        condensation = build_condensation(edges, comp_id, len(sccs))
        build_order = topo_sort(condensation)

        # Should have 4 separate components
        assert len(sccs) == 4

        root_comp = comp_id["root"]
        left_comp = comp_id["left"]
        right_comp = comp_id["right"]
        leaf_comp = comp_id["leaf"]

        root_pos = build_order.index(root_comp)
        left_pos = build_order.index(left_comp)
        right_pos = build_order.index(right_comp)
        leaf_pos = build_order.index(leaf_comp)

        # Leaf must come before left and right
        assert leaf_pos < left_pos
        assert leaf_pos < right_pos
        # Left and right must come before root
        assert left_pos < root_pos
        assert right_pos < root_pos

    def test_multiple_disconnected_sccs(self):
        """Test multiple separate cyclic groups"""
        nodes = ["a1", "a2", "b1", "b2", "b3", "c"]
        edges = [
            ("a1", "a2"),
            ("a2", "a1"),  # First SCC
            ("b1", "b2"),
            ("b2", "b3"),
            ("b3", "b1"),  # Second SCC
            # c is independent
        ]

        sccs, comp_id = tarjan_scc(nodes, edges)
        condensation = build_condensation(edges, comp_id, len(sccs))

        # Should have 3 components
        assert len(sccs) == 3

        # Find SCCs by size
        scc_sizes = sorted([len(scc) for scc in sccs])
        assert scc_sizes == [1, 2, 3]  # Single node, pair, triple

        # All components can be built independently
        assert all(len(neighbors) == 0 for neighbors in condensation.values())

    def test_complex_web_dependencies(self):
        """Test a more complex web-like dependency structure"""
        nodes = [
            "index",
            "navbar",
            "sidebar",
            "main_css",
            "theme_css",
            "logo",
            "bg_image",
        ]
        edges = [
            ("index", "navbar"),
            ("index", "sidebar"),
            ("index", "main_css"),
            ("navbar", "logo"),
            ("navbar", "theme_css"),
            ("sidebar", "theme_css"),
            ("main_css", "theme_css"),
            ("theme_css", "bg_image"),
            (
                "bg_image",
                "main_css",
            ),  # Creates a cycle: main_css -> theme_css -> bg_image -> main_css
        ]

        sccs, comp_id = tarjan_scc(nodes, edges)
        condensation = build_condensation(edges, comp_id, len(sccs))
        build_order = topo_sort(condensation)

        # Should identify the CSS cycle
        assert (
            len(sccs) == 5
        )  # index, navbar, sidebar, logo, {main_css, theme_css, bg_image}

        # Find the CSS SCC
        css_scc = max(sccs, key=len)
        assert len(css_scc) == 3
        assert set(css_scc) == {"main_css", "theme_css", "bg_image"}

        # Verify proper ordering
        index_comp = comp_id["index"]
        css_comp = comp_id["main_css"]  # All CSS components have same comp_id

        css_pos = build_order.index(css_comp)
        index_pos = build_order.index(index_comp)

        # CSS group must be built before index
        assert css_pos < index_pos


class TestResolveBuildOrder:
    def test_mixed_dependencies(self):
        """Test resolve_build_order with both linear dependencies and cycles"""
        nodes = ["home", "api", "cache_a", "cache_b", "config"]
        edges = [
            ("home", "api"),  # home depends on api
            ("api", "cache_a"),  # api depends on cache_a
            ("cache_a", "cache_b"),  # cache_a depends on cache_b (cycle)
            ("cache_b", "cache_a"),  # cache_b depends on cache_a (cycle)
            # config is independent
        ]

        build_groups = resolve_build_order(nodes, edges)

        # Expected: cache cycle first, then config (independent), then api, then home
        expected_groups = [
            {"cache_a", "cache_b"},  # Cycle group (order within may vary)
            {"config"},  # Independent
            {"api"},  # Depends on cache
            {"home"},  # Depends on api
        ]

        actual_groups = [set(group) for group in build_groups]
        assert actual_groups == expected_groups

    def test_unknown_node_in_edge_source(self):
        """Test error when edge references unknown source node"""
        nodes = ["a", "b"]
        edges = [("unknown", "b")]

        with pytest.raises(ValueError, match="Edge references unknown node: unknown"):
            resolve_build_order(nodes, edges)

    def test_unknown_node_in_edge_target(self):
        """Test error when edge references unknown target node"""
        nodes = ["a", "b"]
        edges = [("a", "unknown")]

        with pytest.raises(ValueError, match="Edge references unknown node: unknown"):
            resolve_build_order(nodes, edges)



from datetime import date

import pytest
from django.contrib.contenttypes.models import ContentType
from django.utils import timezone
from freezegun import freeze_time
from wagtail.models import Site

from isekai.contrib.wagtail.loaders import PageLoader
from isekai.contrib.wagtail.transformers import DocumentTransformer, ImageTransformer
from isekai.pipelines import Pipeline
from isekai.types import BlobRef, BlobResource, InMemoryFileProxy, Key, ResourceRef
from tests.testapp.models import ConcreteResource, ReportIndexPage, ReportPage


class TestWagtailImageTransformer:
    def test_transform_image(self):
        transformer = ImageTransformer()

        key = Key(type="url", value="https://example.com/image.png")

        # Create a small PNG image (1x1 red pixel)
        png_data = (
            b"\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x01\x00\x00\x00\x01"
            b"\x08\x02\x00\x00\x00\x90wS\xde\x00\x00\x00\x0cIDATx\x9cc\xf8\x00"
            b"\x00\x00\x03\x00\x01\x00\x00\x00\x00\x18\xdd\x8d\xb4\x1c\x00\x00"
            b"\x00\x00IEND\xaeB`\x82"
        )

        resource = BlobResource(
            mime_type="image/png",
            filename="image.png",
            file_ref=InMemoryFileProxy(content=png_data),
            metadata={"alt_text": "A red pixel"},
        )

        spec = transformer.transform(key, resource)

        assert spec
        assert spec.content_type == "wagtailimages.Image"
        assert spec.attributes == {
            "title": "image.png",
            "file": BlobRef(key),
            "description": "A red pixel",
        }

    def test_disallowed_mime_types(self):
        transformer = ImageTransformer()

        key = Key(type="url", value="https://example.com/image.txt")

        resource = BlobResource(
            mime_type="text/plain",
            filename="image.txt",
            file_ref=InMemoryFileProxy(content=b"Not an image"),
            metadata={},
        )

        spec = transformer.transform(key, resource)

        assert spec is None


class TestWagtailDocumentTransformer:
    def test_transform_document(self):
        transformer = DocumentTransformer()

        key = Key(type="url", value="https://example.com/document.pdf")

        # Create PDF document content
        pdf_data = b"%PDF-1.4\n1 0 obj\n<<\n/Type /Catalog\n/Pages 2 0 R\n>>\nendobj\n"

        resource = BlobResource(
            mime_type="application/pdf",
            filename="document.pdf",
            file_ref=InMemoryFileProxy(content=pdf_data),
            metadata={},
        )

        spec = transformer.transform(key, resource)

        assert spec
        assert spec.content_type == "wagtaildocs.Document"
        assert spec.attributes == {
            "title": "document.pdf",
            "file": BlobRef(key),
        }

    def test_transform_word_document(self):
        transformer = DocumentTransformer()

        key = Key(type="file", value="/path/to/document.docx")

        resource = BlobResource(
            mime_type="application/vnd.openxmlformats-officedocument.wordprocessingml.document",
            filename="report.docx",
            file_ref=InMemoryFileProxy(content=b"Word document content"),
            metadata={},
        )

        spec = transformer.transform(key, resource)

        assert spec
        assert spec.content_type == "wagtaildocs.Document"
        assert spec.attributes == {
            "title": "report.docx",
            "file": BlobRef(key),
        }

    def test_disallowed_mime_types(self):
        transformer = DocumentTransformer()

        key = Key(type="url", value="https://example.com/image.png")

        resource = BlobResource(
            mime_type="image/png",
            filename="image.png",
            file_ref=InMemoryFileProxy(content=b"PNG image data"),
            metadata={},
        )

        spec = transformer.transform(key, resource)

        assert spec is None

    def test_custom_allowed_mime_types(self):
        # Test with custom allowed mime types
        transformer = DocumentTransformer(allowed_mime_types=["text/plain"])

        key = Key(type="file", value="/path/to/text.txt")

        resource = BlobResource(
            mime_type="text/plain",
            filename="text.txt",
            file_ref=InMemoryFileProxy(content=b"Plain text content"),
            metadata={},
        )

        spec = transformer.transform(key, resource)

        assert spec
        assert spec.content_type == "wagtaildocs.Document"

        # Test that PDF is now not allowed
        pdf_resource = BlobResource(
            mime_type="application/pdf",
            filename="document.pdf",
            file_ref=InMemoryFileProxy(content=b"%PDF content"),
            metadata={},
        )

        spec = transformer.transform(key, pdf_resource)
        assert spec is None


@pytest.mark.django_db
@pytest.mark.database_backend
class TestWagtailPageLoader:
    def test_page_loader_creates_parent_child_pages(self):
        """Test that PageLoader creates Wagtail pages"""

        report_index_ct = ContentType.objects.get_for_model(ReportIndexPage)
        report_page_ct = ContentType.objects.get_for_model(ReportPage)

        site_root = Site.objects.get(is_default_site=True).root_page

        report_index_resource = ConcreteResource.objects.create(
            key="url:https://example.com/reports",
            mime_type="application/json",
            data_type="text",
            text_data="unused",
            metadata={},
            target_content_type=report_index_ct,
            target_spec={
                "title": "Reports",
                "intro": "<p>This is the reports index page</p>",
                "slug": "reports",
                "__wagtail_parent_page": site_root.pk,
            },
            status=ConcreteResource.Status.TRANSFORMED,
        )

        report_page_resource = ConcreteResource.objects.create(
            key="url:https://example.com/reports/annual-2023",
            mime_type="application/json",
            data_type="text",
            text_data="unused",
            metadata={},
            target_content_type=report_page_ct,
            target_spec={
                "title": "Annual Report 2023",
                "intro": "<p>Introduction to the annual report</p>",
                "body": "<p>This is the body of the annual report</p>",
                "date": "2023-12-31",
                "slug": "annual-report-2023",
                "__wagtail_parent_page": str(
                    ResourceRef(Key.from_string(report_index_resource.key))
                ),
            },
            status=ConcreteResource.Status.TRANSFORMED,
        )

        report_page_resource.dependencies.add(report_index_resource)

        # Run the pipeline load operation
        now = timezone.now()
        with freeze_time(now):
            pipeline = Pipeline(
                seeders=[],
                extractors=[],
                miners=[],
                transformers=[],
                loaders=[PageLoader()],
            )
            result = pipeline.load()

        # Verify the operation was successful
        assert result.result == "success"

        # Verify the pages were created correctly
        assert ReportIndexPage.objects.filter(title="Reports").exists()
        assert ReportPage.objects.filter(title="Annual Report 2023").exists()

        # Verify parent-child relationship
        created_report_index = ReportIndexPage.objects.get(title="Reports")
        created_report_page = ReportPage.objects.get(title="Annual Report 2023")

        assert created_report_page.get_parent().specific == created_report_index
        assert (
            created_report_index.get_children()
            .filter(pk=created_report_page.pk)
            .exists()
        )

        # Verify page content
        assert created_report_index.intro == "<p>This is the reports index page</p>"
        assert created_report_page.intro == "<p>Introduction to the annual report</p>"
        assert (
            created_report_page.body == "<p>This is the body of the annual report</p>"
        )
        assert created_report_page.date == date(2023, 12, 31)

        # Verify URLs
        assert created_report_index.slug == "reports"
        assert created_report_page.slug == "annual-report-2023"

        # Verify resources are marked as loaded
        report_index_resource.refresh_from_db()
        report_page_resource.refresh_from_db()

        assert report_index_resource.status == ConcreteResource.Status.LOADED
        assert report_page_resource.status == ConcreteResource.Status.LOADED
        assert report_index_resource.target_object_id == str(created_report_index.pk)
        assert report_page_resource.target_object_id == str(created_report_page.pk)
        assert report_index_resource.loaded_at == now
        assert report_page_resource.loaded_at == now

    def test_page_loader_creates_complex_page_hierarchy(self):
        """Test PageLoader with multiple depths, siblings, and complex hierarchy."""

        report_index_ct = ContentType.objects.get_for_model(ReportIndexPage)
        report_page_ct = ContentType.objects.get_for_model(ReportPage)

        site_root = Site.objects.get(is_default_site=True).root_page

        # Level 1: Top-level sections (siblings under site root)
        reports_section = ConcreteResource.objects.create(
            key="url:https://example.com/reports",
            target_content_type=report_index_ct,
            target_spec={
                "title": "Reports Section",
                "intro": "<p>All company reports</p>",
                "slug": "reports",
                "__wagtail_parent_page": site_root.pk,
            },
            status=ConcreteResource.Status.TRANSFORMED,
        )

        news_section = ConcreteResource.objects.create(
            key="url:https://example.com/news",
            target_content_type=report_index_ct,
            target_spec={
                "title": "News Section",
                "intro": "<p>Company news and updates</p>",
                "slug": "news",
                "__wagtail_parent_page": site_root.pk,
            },
            status=ConcreteResource.Status.TRANSFORMED,
        )

        # Level 2: Sub-sections under Reports (siblings)
        annual_reports = ConcreteResource.objects.create(
            key="url:https://example.com/reports/annual",
            target_content_type=report_index_ct,
            target_spec={
                "title": "Annual Reports",
                "intro": "<p>Yearly financial reports</p>",
                "slug": "annual",
                "__wagtail_parent_page": str(
                    ResourceRef(Key.from_string(reports_section.key))
                ),
            },
            status=ConcreteResource.Status.TRANSFORMED,
        )

        quarterly_reports = ConcreteResource.objects.create(
            key="url:https://example.com/reports/quarterly",
            target_content_type=report_index_ct,
            target_spec={
                "title": "Quarterly Reports",
                "intro": "<p>Quarterly business updates</p>",
                "slug": "quarterly",
                "__wagtail_parent_page": str(
                    ResourceRef(Key.from_string(reports_section.key))
                ),
            },
            status=ConcreteResource.Status.TRANSFORMED,
        )

        # Level 2: Pages under News (siblings)
        press_releases = ConcreteResource.objects.create(
            key="url:https://example.com/news/press",
            target_content_type=report_page_ct,
            target_spec={
                "title": "Press Releases",
                "intro": "<p>Official company announcements</p>",
                "body": "<p>Latest press releases and media coverage</p>",
                "date": "2023-01-15",
                "slug": "press-releases",
                "__wagtail_parent_page": str(
                    ResourceRef(Key.from_string(news_section.key))
                ),
            },
            status=ConcreteResource.Status.TRANSFORMED,
        )

        company_blog = ConcreteResource.objects.create(
            key="url:https://example.com/news/blog",
            target_content_type=report_page_ct,
            target_spec={
                "title": "Company Blog",
                "intro": "<p>Insights from our team</p>",
                "body": "<p>Regular updates and insights from our company</p>",
                "date": "2023-02-01",
                "slug": "company-blog",
                "__wagtail_parent_page": str(
                    ResourceRef(Key.from_string(news_section.key))
                ),
            },
            status=ConcreteResource.Status.TRANSFORMED,
        )

        # Level 3: Individual reports under Annual Reports (siblings)
        annual_2023 = ConcreteResource.objects.create(
            key="url:https://example.com/reports/annual/2023",
            target_content_type=report_page_ct,
            target_spec={
                "title": "Annual Report 2023",
                "intro": "<p>Financial performance for 2023</p>",
                "body": "<p>Detailed financial analysis and company performance for fiscal year 2023</p>",
                "date": "2023-12-31",
                "slug": "annual-report-2023",
                "__wagtail_parent_page": str(
                    ResourceRef(Key.from_string(annual_reports.key))
                ),
            },
            status=ConcreteResource.Status.TRANSFORMED,
        )

        annual_2022 = ConcreteResource.objects.create(
            key="url:https://example.com/reports/annual/2022",
            target_content_type=report_page_ct,
            target_spec={
                "title": "Annual Report 2022",
                "intro": "<p>Financial performance for 2022</p>",
                "body": "<p>Comprehensive review of company achievements in fiscal year 2022</p>",
                "date": "2022-12-31",
                "slug": "annual-report-2022",
                "__wagtail_parent_page": str(
                    ResourceRef(Key.from_string(annual_reports.key))
                ),
            },
            status=ConcreteResource.Status.TRANSFORMED,
        )

        # Level 3: Individual reports under Quarterly Reports (siblings)
        q4_2023 = ConcreteResource.objects.create(
            key="url:https://example.com/reports/quarterly/q4-2023",
            target_content_type=report_page_ct,
            target_spec={
                "title": "Q4 2023 Report",
                "intro": "<p>Fourth quarter results</p>",
                "body": "<p>Q4 2023 quarterly business performance and highlights</p>",
                "date": "2023-12-31",
                "slug": "q4-2023-report",
                "__wagtail_parent_page": str(
                    ResourceRef(Key.from_string(quarterly_reports.key))
                ),
            },
            status=ConcreteResource.Status.TRANSFORMED,
        )

        q3_2023 = ConcreteResource.objects.create(
            key="url:https://example.com/reports/quarterly/q3-2023",
            target_content_type=report_page_ct,
            target_spec={
                "title": "Q3 2023 Report",
                "intro": "<p>Third quarter results</p>",
                "body": "<p>Q3 2023 quarterly performance analysis and market outlook</p>",
                "date": "2023-09-30",
                "slug": "q3-2023-report",
                "__wagtail_parent_page": str(
                    ResourceRef(Key.from_string(quarterly_reports.key))
                ),
            },
            status=ConcreteResource.Status.TRANSFORMED,
        )

        # Set up dependency relationships (children depend on parents)
        annual_reports.dependencies.add(reports_section)
        quarterly_reports.dependencies.add(reports_section)
        press_releases.dependencies.add(news_section)
        company_blog.dependencies.add(news_section)
        annual_2023.dependencies.add(annual_reports)
        annual_2022.dependencies.add(annual_reports)
        q4_2023.dependencies.add(quarterly_reports)
        q3_2023.dependencies.add(quarterly_reports)

        # Run the pipeline load operation
        now = timezone.now()
        with freeze_time(now):
            pipeline = Pipeline(
                seeders=[],
                extractors=[],
                miners=[],
                transformers=[],
                loaders=[PageLoader()],
            )
            result = pipeline.load()

        # Verify the operation was successful
        assert result.result == "success"

        # Verify all pages were created
        assert ReportIndexPage.objects.filter(title="Reports Section").exists()
        assert ReportIndexPage.objects.filter(title="News Section").exists()
        assert ReportIndexPage.objects.filter(title="Annual Reports").exists()
        assert ReportIndexPage.objects.filter(title="Quarterly Reports").exists()
        assert ReportPage.objects.filter(title="Press Releases").exists()
        assert ReportPage.objects.filter(title="Company Blog").exists()
        assert ReportPage.objects.filter(title="Annual Report 2023").exists()
        assert ReportPage.objects.filter(title="Annual Report 2022").exists()
        assert ReportPage.objects.filter(title="Q4 2023 Report").exists()
        assert ReportPage.objects.filter(title="Q3 2023 Report").exists()

        # Get created pages for relationship verification
        reports_section_page = ReportIndexPage.objects.get(title="Reports Section")
        news_section_page = ReportIndexPage.objects.get(title="News Section")
        annual_reports_page = ReportIndexPage.objects.get(title="Annual Reports")
        quarterly_reports_page = ReportIndexPage.objects.get(title="Quarterly Reports")
        press_releases_page = ReportPage.objects.get(title="Press Releases")
        company_blog_page = ReportPage.objects.get(title="Company Blog")
        annual_2023_page = ReportPage.objects.get(title="Annual Report 2023")
        annual_2022_page = ReportPage.objects.get(title="Annual Report 2022")
        q4_2023_page = ReportPage.objects.get(title="Q4 2023 Report")
        q3_2023_page = ReportPage.objects.get(title="Q3 2023 Report")

        # Verify Level 1: Top-level sections are children of site root
        assert reports_section_page.get_parent() == site_root
        assert news_section_page.get_parent() == site_root

        # Verify Level 2: Sub-sections have correct parents
        assert annual_reports_page.get_parent().specific == reports_section_page
        assert quarterly_reports_page.get_parent().specific == reports_section_page
        assert press_releases_page.get_parent().specific == news_section_page
        assert company_blog_page.get_parent().specific == news_section_page

        # Verify Level 3: Individual reports have correct parents
        assert annual_2023_page.get_parent().specific == annual_reports_page
        assert annual_2022_page.get_parent().specific == annual_reports_page
        assert q4_2023_page.get_parent().specific == quarterly_reports_page
        assert q3_2023_page.get_parent().specific == quarterly_reports_page

        # Verify sibling relationships at Level 1
        # Refresh site_root to get the latest children
        site_root.refresh_from_db()
        level1_children = site_root.get_children().specific()
        level1_titles = {page.title for page in level1_children}
        assert "Reports Section" in level1_titles
        assert "News Section" in level1_titles

        # Verify sibling relationships at Level 2 under Reports
        reports_children = reports_section_page.get_children().specific()
        reports_children_titles = {page.title for page in reports_children}
        assert reports_children_titles == {"Annual Reports", "Quarterly Reports"}

        # Verify sibling relationships at Level 2 under News
        news_children = news_section_page.get_children().specific()
        news_children_titles = {page.title for page in news_children}
        assert news_children_titles == {"Press Releases", "Company Blog"}

        # Verify sibling relationships at Level 3 under Annual Reports
        annual_children = annual_reports_page.get_children().specific()
        annual_children_titles = {page.title for page in annual_children}
        assert annual_children_titles == {"Annual Report 2023", "Annual Report 2022"}

        # Verify sibling relationships at Level 3 under Quarterly Reports
        quarterly_children = quarterly_reports_page.get_children().specific()
        quarterly_children_titles = {page.title for page in quarterly_children}
        assert quarterly_children_titles == {"Q4 2023 Report", "Q3 2023 Report"}

        # Verify depths are correct (Wagtail manages this automatically)
        assert reports_section_page.depth == site_root.depth + 1
        assert annual_reports_page.depth == reports_section_page.depth + 1
        assert annual_2023_page.depth == annual_reports_page.depth + 1

        # Verify all resources are marked as loaded
        all_resources = [
            reports_section,
            news_section,
            annual_reports,
            quarterly_reports,
            press_releases,
            company_blog,
            annual_2023,
            annual_2022,
            q4_2023,
            q3_2023,
        ]

        for resource in all_resources:
            resource.refresh_from_db()
            assert resource.status == ConcreteResource.Status.LOADED
            assert resource.loaded_at == now
            assert resource.target_object_id is not None



import os

import django
import pytest

# Configure Django before importing models
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "tests.settings")
django.setup()


@pytest.fixture(autouse=True)
def _media_tmpdir(settings, tmp_path):
    settings.MEDIA_ROOT = tmp_path


@pytest.fixture(autouse=True, scope="session")
def _setup_wagtail_initial_data(django_db_setup, django_db_blocker):
    """Create Wagtail's initial data (root page, site, collection) when migrations are disabled."""
    with django_db_blocker.unblock():
        from django.conf import settings
        from django.contrib.contenttypes.models import ContentType
        from wagtail.models import Page, Site
        from wagtail.models.i18n import Locale
        from wagtail.models.media import Collection

        # Create default locale
        if not Locale.objects.exists():
            Locale.objects.get_or_create(
                language_code=settings.LANGUAGE_CODE,
                defaults={"language_code": "en"},
            )

        # Create root collection if it doesn't exist
        if not Collection.objects.exists():
            Collection.add_root(name="Root")

        # Create root page if it doesn't exist
        if not Page.objects.exists():
            page_content_type = ContentType.objects.get_for_model(Page)
            root_page = Page.add_root(  # type: ignore[attr-defined]
                title="Root",
                slug="root",
                content_type=page_content_type,
            )

            # Create default site
            Site.objects.get_or_create(
                hostname="localhost",
                defaults={
                    "root_page": root_page,
                    "is_default_site": True,
                    "site_name": "Test Site",
                },
            )



import pytest
from django.contrib.contenttypes.models import ContentType
from django.core.files.base import ContentFile
from django.utils import timezone
from freezegun import freeze_time

from isekai.pipelines import Pipeline, get_django_pipeline
from isekai.transformers import BaseTransformer
from isekai.types import BlobRef, Key, ModelRef, ResourceRef, Spec, TextResource
from tests.testapp.models import ConcreteResource


@pytest.mark.django_db
class TestTransform:
    def test_transform_saves_specs(self):
        resource = ConcreteResource.objects.create(
            key="url:https://example.com/image.png",
            data_type="blob",
            mime_type="image/png",
            metadata={"alt_text": "A sample image"},
            status=ConcreteResource.Status.MINED,
        )
        resource.blob_data.save("image.png", ContentFile(b"fake image data"))

        now = timezone.now()
        with freeze_time(now):
            pipeline = get_django_pipeline()
            pipeline.transform()

        resource.refresh_from_db()

        assert resource.target_content_type == ContentType.objects.get(
            app_label="wagtailimages", model="image"
        )
        assert resource.target_spec == {
            "title": "image.png",
            "file": "isekai-blob-ref:\\url:https://example.com/image.png",
            "description": "A sample image",
        }

        assert resource.status == ConcreteResource.Status.TRANSFORMED
        assert resource.transformed_at == now

        # Verify dependencies are set based on refs in the spec
        dependencies = list(resource.dependencies.all())
        assert len(dependencies) == 1
        assert dependencies[0].key == "url:https://example.com/image.png"
        assert (
            dependencies[0] == resource
        )  # Should reference itself since it's a BlobRef to itself

    def test_transform_handles_transformer_chaining(self):
        """Test that transform operation handles transformer chaining correctly."""

        # Create a resource with foo/bar mime type that should be handled by FooBarTransformer
        resource = ConcreteResource.objects.create(
            key="foo:bar-transform-test",
            data_type="text",
            mime_type="foo/bar",
            text_data="some foo bar text",
            metadata={"source": "test"},
            status=ConcreteResource.Status.MINED,
        )

        now = timezone.now()
        with freeze_time(now):
            pipeline = get_django_pipeline()
            pipeline.transform()

        resource.refresh_from_db()

        # Should be transformed by FooBarTransformer
        assert resource.target_content_type == ContentType.objects.get(
            app_label="auth", model="user"
        )
        assert resource.target_spec == {
            "username": "foobar_user",
            "email": "foo@bar.com",
        }

        assert resource.status == ConcreteResource.Status.TRANSFORMED
        assert resource.transformed_at == now

    def test_transform_is_idempotent(self):
        """Test that running transform multiple times doesn't re-transform already transformed resources."""
        resource = ConcreteResource.objects.create(
            key="url:https://example.com/test-image.png",
            data_type="blob",
            mime_type="image/png",
            metadata={"alt_text": "Test image"},
            status=ConcreteResource.Status.MINED,
        )
        resource.blob_data.save("test-image.png", ContentFile(b"test image data"))

        # First transform operation
        now = timezone.now()
        with freeze_time(now):
            pipeline = get_django_pipeline()
            pipeline.transform()

        # Verify resource was transformed
        resource.refresh_from_db()
        assert resource.status == ConcreteResource.Status.TRANSFORMED
        assert resource.transformed_at == now
        original_target_spec = resource.target_spec.copy()
        original_content_type_id = resource.target_content_type_id

        # Second transform operation - should not process already transformed resources
        later = now + timezone.timedelta(hours=1)
        with freeze_time(later):
            pipeline = get_django_pipeline()
            pipeline.transform()  # Should be no-op

        # Verify resource state unchanged
        resource.refresh_from_db()
        assert resource.status == ConcreteResource.Status.TRANSFORMED
        assert resource.transformed_at == now  # Timestamp should not change
        assert (
            resource.target_spec == original_target_spec
        )  # Spec should remain the same
        assert (
            resource.target_content_type_id == original_content_type_id
        )  # Content type should remain the same

    def test_no_transformer_found_for_resource(self):
        """Test that resource remains unchanged if no transformer is found."""
        resource = ConcreteResource.objects.create(
            key="url:https://example.com/unknown-file.xyz",
            data_type="blob",
            mime_type="application/xyz",  # Unsupported mime type
            metadata={},
            status=ConcreteResource.Status.MINED,
        )
        resource.blob_data.save("unknown-file.xyz", ContentFile(b"unknown data"))

        now = timezone.now()
        with freeze_time(now):
            pipeline = get_django_pipeline()
            pipeline.transform()

        resource.refresh_from_db()

        # Resource should remain in MINED status with no target spec/content type
        assert resource.status == ConcreteResource.Status.MINED
        assert resource.transformed_at is None
        assert resource.target_spec is None
        assert resource.target_content_type is None
        assert (
            resource.last_error
            == "TransformError: No transformer could handle the resource"
        )

    def test_transform_sets_dependencies_from_spec_refs(self):
        """Test that transform operation sets dependencies based on refs in the transformed spec."""
        # Test with FooBarTransformer which doesn't create refs - should have no dependencies
        resource_no_deps = ConcreteResource.objects.create(
            key="foo:test-no-deps",
            data_type="text",
            mime_type="foo/bar",
            text_data="some foo bar text",
            metadata={"source": "test"},
            status=ConcreteResource.Status.MINED,
        )

        pipeline = get_django_pipeline()
        pipeline.transform()

        resource_no_deps.refresh_from_db()

        # FooBarTransformer creates a spec with no refs, so no dependencies
        assert resource_no_deps.status == ConcreteResource.Status.TRANSFORMED
        dependencies = list(resource_no_deps.dependencies.all())
        assert len(dependencies) == 0

    def test_transform_sets_multiple_dependencies_from_spec_refs(self):
        """Test that transform operation correctly sets multiple dependencies from spec refs."""

        # Create dependency resources that will be referenced
        image_resource = ConcreteResource.objects.create(
            key="url:https://example.com/hero-image.jpg",
            status=ConcreteResource.Status.LOADED,
        )

        author_resource = ConcreteResource.objects.create(
            key="gen:author-456",
            status=ConcreteResource.Status.LOADED,
        )

        attachment_resource = ConcreteResource.objects.create(
            key="file:document.pdf",
            status=ConcreteResource.Status.LOADED,
        )

        # Create main resource to transform
        main_resource = ConcreteResource.objects.create(
            key="url:https://example.com/article.html",
            data_type="text",
            mime_type="application/x-test-multi-ref",
            text_data="Test article content",
            metadata={"title": "Test Article"},
            status=ConcreteResource.Status.MINED,
        )

        # Create a custom transformer that produces a spec with multiple refs
        class MultiRefTransformer(BaseTransformer):
            def transform(self, key: Key, resource):
                if resource.mime_type != "application/x-test-multi-ref":
                    return None

                assert isinstance(resource, TextResource)

                return Spec(
                    content_type="testapp.Article",
                    attributes={
                        "title": resource.metadata.get("title", "Default"),
                        "content": resource.text,
                        "hero_image": BlobRef(
                            Key(type="url", value="https://example.com/hero-image.jpg")
                        ),
                        "author": ResourceRef(Key(type="gen", value="author-456")),
                        "category": ModelRef("testapp.Category", pk=1),
                        "attachment": BlobRef(Key(type="file", value="document.pdf")),
                        "metadata": {
                            "related_image": BlobRef(
                                Key(
                                    type="url",
                                    value="https://example.com/hero-image.jpg",
                                )
                            ),  # Duplicate ref
                            "secondary_author": ResourceRef(
                                Key(type="gen", value="author-456")
                            ),  # Duplicate ref
                        },
                        # Non-ref values (should be ignored for dependencies)
                        "published": True,
                        "tags": ["tech", "news"],
                    },
                )

        # Add content type for our test
        ContentType.objects.get_or_create(
            app_label="testapp",
            model="article",
        )

        # Create pipeline with custom transformer
        pipeline = Pipeline(
            seeders=[],
            extractors=[],
            miners=[],
            transformers=[MultiRefTransformer()],
            loaders=[],
        )

        now = timezone.now()
        with freeze_time(now):
            pipeline.transform()

        main_resource.refresh_from_db()

        # Verify resource was transformed
        assert main_resource.status == ConcreteResource.Status.TRANSFORMED
        assert main_resource.transformed_at == now

        # Verify the spec was created correctly with refs
        expected_spec = {
            "title": "Test Article",
            "content": "Test article content",
            "hero_image": "isekai-blob-ref:\\url:https://example.com/hero-image.jpg",
            "author": "isekai-resource-ref:\\gen:author-456",
            "category": "isekai-model-ref:\\testapp.Category?pk=1",
            "attachment": "isekai-blob-ref:\\file:document.pdf",
            "metadata": {
                "related_image": "isekai-blob-ref:\\url:https://example.com/hero-image.jpg",
                "secondary_author": "isekai-resource-ref:\\gen:author-456",
            },
            "published": True,
            "tags": ["tech", "news"],
        }
        assert main_resource.target_spec == expected_spec

        # Verify dependencies are set correctly (unique refs only, no duplicates)
        # Note: ModelRef doesn't create dependencies since it references external DB objects
        dependencies = list(main_resource.dependencies.all())
        dependency_keys = {dep.key for dep in dependencies}

        expected_dependency_keys = {
            "url:https://example.com/hero-image.jpg",  # BlobRef (appears twice but should be unique)
            "gen:author-456",  # ResourceRef (appears twice but should be unique)
            "file:document.pdf",  # BlobRef
        }

        assert len(dependencies) == 3  # Should have exactly 3 unique dependencies
        assert dependency_keys == expected_dependency_keys

        # Verify the actual dependency objects exist
        assert image_resource in dependencies
        assert author_resource in dependencies
        assert attachment_resource in dependencies

    def test_transform_fails_with_invalid_refs(self):
        """Test that transform operation fails when specs reference non-existent resources."""

        # Create main resource to transform
        main_resource = ConcreteResource.objects.create(
            key="url:https://example.com/invalid-refs.html",
            data_type="text",
            mime_type="application/x-test-invalid-refs",
            text_data="Test content with invalid refs",
            metadata={"title": "Invalid Refs Test"},
            status=ConcreteResource.Status.MINED,
        )

        # Create a transformer that produces a spec with refs to non-existent resources
        class InvalidRefTransformer(BaseTransformer):
            def transform(self, key: Key, resource):
                if resource.mime_type != "application/x-test-invalid-refs":
                    return None

                assert isinstance(resource, TextResource)

                return Spec(
                    content_type="testapp.Article",
                    attributes={
                        "title": resource.metadata.get("title", "Default"),
                        "content": resource.text,
                        # Refs to resources that don't exist - this should cause an error
                        "nonexistent_image": BlobRef(
                            Key(type="url", value="https://example.com/missing.jpg")
                        ),
                    },
                )

        # Add content type for our test
        ContentType.objects.get_or_create(
            app_label="testapp",
            model="article",
        )

        # Create pipeline with transformer that references non-existent resources
        pipeline = Pipeline(
            seeders=[],
            extractors=[],
            miners=[],
            transformers=[InvalidRefTransformer()],
            loaders=[],
        )

        # The transform operation should catch the invalid refs and handle them gracefully
        pipeline.transform()

        # Verify that the resource remains in MINED status due to the error
        main_resource.refresh_from_db()
        assert main_resource.status == ConcreteResource.Status.MINED

        # Verify that the error was captured
        assert main_resource.last_error == "TransformError: Invalid refs found in spec"



import time

from django.core.management.base import BaseCommand
from rich import box
from rich.console import Console
from rich.table import Table

import isekai
from isekai.pipelines import get_django_pipeline
from isekai.types import Operation, OperationResult
from isekai.utils.progress import LiveProgressLogger


def get_result_display(result: str) -> str:
    """Map result string to display status."""
    if result == "success":
        return "OK"
    elif result == "partial_success":
        return "WARN"
    else:
        return "ERROR"

    return "UNKNOWN"


class Command(BaseCommand):
    help = "Run isekai ETL operations with live progress display"

    console: Console

    def add_arguments(self, parser):
        parser.add_argument(
            "--no-input",
            action="store_true",
            help="Skip interactive prompts and run automatically",
        )

    def handle(self, *args, **options):
        console = Console(file=self.stdout)  # type: ignore

        # Create the header with Unicode underline
        version = isekai.__version__
        header = f"ISEKAI v{version} (Your data in another world)"
        underline = "─" * len(header)

        # Print the header
        console.print(header, style="bold")
        console.print(underline)
        console.print("")  # Empty line

        # Display pipeline configuration
        pipeline = get_django_pipeline()

        try:
            pipeline_config = pipeline.get_configuration()
        except Exception as e:
            console.print(f"Error loading pipeline configuration: {e}")
            return

        # Print Pipeline header separately (not italicized)
        console.print("Pipeline")

        # Create table without title
        table = Table(show_header=True, header_style="bold", box=box.SQUARE)
        table.add_column("Stage", style="cyan", no_wrap=True)
        table.add_column("Processors", style="white")

        # Add rows for each stage
        for stage, processors in pipeline_config.items():
            if processors:
                # Add first processor in the stage
                table.add_row(stage, processors[0])
                # Add remaining processors with empty stage column
                for processor in processors[1:]:
                    table.add_row("", processor)
            else:
                # No processors for this stage
                table.add_row(stage, "[dim]None[/dim]")

        # Display table
        console.print(table)
        console.print()

        # Ask for confirmation unless --no-input is specified
        if not options.get("no_input", False):
            response = input("Start pipeline? [y/N]: ").lower().strip()
            if response not in ["y", "yes"]:
                console.print("Pipeline cancelled.")
                return

        # Run the pipeline
        console.print()
        pipeline_start = time.time()
        progress_logger = LiveProgressLogger()

        results = []

        # Seed
        seed_result = self.execute_step(
            console,
            progress_logger,
            "Seeding",
            pipeline.seed,
        )
        if seed_result:
            results.append(seed_result)

        # Extract and Mine loop
        newly_seeded_count = 1  # Initialize to enter the loop
        while newly_seeded_count > 0:
            # Extract
            execute_result = self.execute_step(
                console,
                progress_logger,
                "Extracting",
                pipeline.extract,
            )

            if execute_result:
                results.append(execute_result)

            # Mine
            mine_result = self.execute_step(
                console,
                progress_logger,
                "Mining",
                pipeline.mine,
            )

            if mine_result:
                newly_seeded_count = mine_result.metadata.get("newly_seeded_count", 0)
                results.append(mine_result)
            else:
                # If mining failed, stop the loop
                newly_seeded_count = 0

        # Transform
        transform_result = self.execute_step(
            console,
            progress_logger,
            "Transforming",
            pipeline.transform,
        )
        if transform_result:
            results.append(transform_result)

        # Load
        load_result = self.execute_step(
            console,
            progress_logger,
            "Loading",
            pipeline.load,
        )
        if load_result:
            results.append(load_result)

        # Calculate total time
        total_time = time.time() - pipeline_start

        # Print completion message
        console.print(f"[[green]DONE[/green]] Pipeline finished in {total_time:.1f}s")
        console.print()

        object_stats = {}
        if load_result and load_result.metadata:
            object_stats = load_result.metadata.get("object_stats", {})

        # Print summary table
        console.print("Summary")
        summary_table = Table(show_header=True, header_style="bold", box=box.SQUARE)
        summary_table.add_column("Object Type", style="cyan", no_wrap=True)
        summary_table.add_column("Count", style="white")

        # Add rows for each object type
        for obj_type, count in object_stats.items():
            summary_table.add_row(obj_type, str(count))

        # Display table
        console.print(summary_table)
        console.print()

        # Calculate overall status
        has_errors = any(r.result == "failure" for r in results)
        has_warnings = any(r.result == "partial_success" for r in results)

        if has_errors:
            status = "[red]FAILED[/red]"
        elif has_warnings:
            status = "[yellow]COMPLETED WITH WARNINGS[/yellow]"
        else:
            status = "[green]COMPLETED SUCCESSFULLY[/green]"

        console.print(f"Result: {status}")

    def execute_step(
        self,
        console: Console,
        progress_logger: LiveProgressLogger,
        step_name: str,
        step_func: Operation,
    ) -> OperationResult | None:
        messages = []
        with progress_logger.task(step_name) as task_manager:
            try:
                result = step_func()
                messages = result.messages
                task_manager.set_status(get_result_display(result.result))
            except Exception as e:
                result = None
                messages = [f"Error: {e}"]
                task_manager.set_status("ERROR")

        for message in messages:
            console.print(f"  {message}")

        console.print("")  # Empty line between operations

        return result



from isekai.utils.core import get_resource_model
from tests.testapp.models import ConcreteResource


class TestGetResourceModel:
    def test_get_resource_model_returns_concrete_resource(self):
        """Test that get_resource_model returns the concrete resource class."""
        model = get_resource_model()
        assert model is ConcreteResource



from isekai.types import (
    BlobRef,
    Key,
    ModelRef,
    ResourceRef,
    Spec,
    ref,
)
from tests.test_models import pytest


class TestKey:
    def test_key(self):
        key = Key(type="test", value="123")
        assert str(key) == "test:123"

        # Key from string
        key_from_string = Key.from_string("test:123")
        assert key_from_string == key

    def test_invalid_key_string(self):
        # Invalid format should raise ValueError
        with pytest.raises(ValueError):
            Key.from_string("invalid-key")

        # No value should raise ValueError
        with pytest.raises(ValueError):
            Key.from_string("test:")


class TestRefs:
    def test_blob_ref(self):
        key = Key(type="blob", value="456")

        # Key to BlobRef
        blob_ref = BlobRef(key)
        assert str(blob_ref) == "isekai-blob-ref:\\blob:456"

        # BlobRef to Key
        blob_ref = BlobRef.from_string("isekai-blob-ref:\\blob:456")
        assert blob_ref.key == key

    def test_blob_ref_invalid_string(self):
        # Invalid format should raise ValueError
        with pytest.raises(ValueError):
            BlobRef.from_string("invalid-blob-string")

        # Invalid prefix should raise ValueError
        with pytest.raises(ValueError):
            BlobRef.from_string("blob:\\test:456")

    def test_resource_ref_basic(self):
        key = Key(type="user", value="123")

        # Basic ResourceRef without attribute access
        ref = ResourceRef(key)
        assert str(ref) == "isekai-resource-ref:\\user:123"

        # ResourceRef from string
        ref = ResourceRef.from_string("isekai-resource-ref:\\user:123")
        assert ref.key == key
        assert ref.ref_attr_path == ()

    def test_resource_ref_with_pk_attribute(self):
        key = Key(type="user", value="123")

        # ResourceRef with .pk access
        ref = ResourceRef(key).pk
        assert str(ref) == "isekai-resource-ref:\\user:123::pk"

        # ResourceRef from string with pk
        ref = ResourceRef.from_string("isekai-resource-ref:\\user:123::pk")
        assert ref.key == key
        assert ref.ref_attr_path == ("pk",)

    def test_resource_ref_with_chained_attributes(self):
        key = Key(type="user", value="123")

        # ResourceRef with chained attribute access
        ref = ResourceRef(key).group.name
        assert str(ref) == "isekai-resource-ref:\\user:123::group.name"

        # ResourceRef from string with chained attributes
        ref = ResourceRef.from_string("isekai-resource-ref:\\user:123::group.name")
        assert ref.key == key
        assert ref.ref_attr_path == ("group", "name")

    def test_resource_ref_with_deep_chaining(self):
        key = Key(type="article", value="456")

        # ResourceRef with deep chaining
        ref = ResourceRef(key).author.group.name.slug
        assert str(ref) == "isekai-resource-ref:\\article:456::author.group.name.slug"

        # ResourceRef from string with deep chaining
        ref = ResourceRef.from_string(
            "isekai-resource-ref:\\article:456::author.group.name.slug"
        )
        assert ref.key == key
        assert ref.ref_attr_path == ("author", "group", "name", "slug")

    def test_resource_ref_invalid_string(self):
        # Invalid format should raise ValueError
        with pytest.raises(ValueError):
            ResourceRef.from_string("invalid-string")

        # Invalid prefix should raise ValueError
        with pytest.raises(ValueError):
            ResourceRef.from_string("resource:\\test:123")

    def test_model_ref_basic(self):
        # Basic ModelRef with single kwarg
        ref = ModelRef("testapp.Author", pk=42)
        assert str(ref) == "isekai-model-ref:\\testapp.Author?pk=42"

        # ModelRef from string
        ref = ModelRef.from_string("isekai-model-ref:\\testapp.Author?pk=42")
        assert ref.ref_content_type == "testapp.Author"
        assert ref.ref_lookup_kwargs == {"pk": "42"}
        assert ref.ref_attr_path == ()

    def test_model_ref_multiple_kwargs(self):
        # ModelRef with multiple kwargs
        ref = ModelRef("auth.User", email="test@example.com", is_active=True)
        ref_str = str(ref)
        assert ref_str.startswith("isekai-model-ref:\\auth.User?")
        # Query params can be in any order, so check both contain the same params
        assert "email=test%40example.com" in ref_str
        assert "is_active=True" in ref_str

        # ModelRef from string with multiple params
        ref = ModelRef.from_string(
            "isekai-model-ref:\\auth.User?email=test%40example.com&is_active=True"
        )
        assert ref.ref_content_type == "auth.User"
        assert ref.ref_lookup_kwargs == {
            "email": "test@example.com",
            "is_active": "True",
        }

    def test_model_ref_with_attribute_access(self):
        # ModelRef with attribute access
        ref = ModelRef("testapp.Author", pk=42).group.name
        assert str(ref) == "isekai-model-ref:\\testapp.Author?pk=42::group.name"

        # ModelRef from string with attributes
        ref = ModelRef.from_string(
            "isekai-model-ref:\\testapp.Author?pk=42::group.name"
        )
        assert ref.ref_content_type == "testapp.Author"
        assert ref.ref_lookup_kwargs == {"pk": "42"}
        assert ref.ref_attr_path == ("group", "name")

    def test_model_ref_with_deep_chaining(self):
        # ModelRef with deep attribute chaining
        ref = ModelRef("testapp.Article", slug="my-article").author.group.name
        ref_str = str(ref)
        assert ref_str.startswith("isekai-model-ref:\\testapp.Article?")
        assert "slug=my-article" in ref_str
        assert ref_str.endswith("::author.group.name")

        # ModelRef from string with deep chaining
        ref = ModelRef.from_string(
            "isekai-model-ref:\\testapp.Article?slug=my-article::author.group.name"
        )
        assert ref.ref_content_type == "testapp.Article"
        assert ref.ref_lookup_kwargs == {"slug": "my-article"}
        assert ref.ref_attr_path == ("author", "group", "name")

    def test_model_ref_with_pk_attribute(self):
        # ModelRef with .pk access
        ref = ModelRef("testapp.Author", email="test@example.com").pk
        assert (
            str(ref) == "isekai-model-ref:\\testapp.Author?email=test%40example.com::pk"
        )

        # ModelRef from string with pk
        ref = ModelRef.from_string(
            "isekai-model-ref:\\testapp.Author?email=test%40example.com::pk"
        )
        assert ref.ref_content_type == "testapp.Author"
        assert ref.ref_lookup_kwargs == {"email": "test@example.com"}
        assert ref.ref_attr_path == ("pk",)

    def test_model_ref_invalid_string(self):
        # Invalid format should raise ValueError
        with pytest.raises(ValueError):
            ModelRef.from_string("invalid-string")

        # Invalid prefix should raise ValueError
        with pytest.raises(ValueError):
            ModelRef.from_string("model:\\testapp.Author?pk=42")


class TestSpec:
    def test_to_dict(self):
        spec = Spec(
            content_type="foo.Bar",
            attributes={
                "title": "Test Title",
                "image": BlobRef(
                    Key(type="url", value="https://example.com/image.png")
                ),
                "description": "A sample description",
                "call_to_action": ResourceRef(
                    Key(type="gen", value="call_to_action_123")
                ).pk,
                "child_object": {
                    "pk": ResourceRef(Key(type="gen", value="child_object_456")).pk,
                    "name": "Child Object Name",
                },
            },
        )

        expected_dict = {
            "content_type": "foo.Bar",
            "attributes": {
                "title": "Test Title",
                "image": "isekai-blob-ref:\\url:https://example.com/image.png",
                "description": "A sample description",
                "call_to_action": "isekai-resource-ref:\\gen:call_to_action_123::pk",
                "child_object": {
                    "pk": "isekai-resource-ref:\\gen:child_object_456::pk",
                    "name": "Child Object Name",
                },
            },
        }

        assert spec.to_dict() == expected_dict

    def test_to_dict_with_lists(self):
        spec = Spec(
            content_type="foo.ListContainer",
            attributes={
                "images": [
                    BlobRef(Key(type="file", value="image1.jpg")),
                    BlobRef(Key(type="file", value="image2.jpg")),
                ],
                "references": [
                    ResourceRef(Key(type="gen", value="ref1")).pk,
                    ResourceRef(Key(type="gen", value="ref2")).pk,
                ],
                "mixed_list": [
                    "string_value",
                    42,
                    ResourceRef(Key(type="gen", value="mixed_ref")).pk,
                    {"nested_ref": BlobRef(Key(type="url", value="nested.png"))},
                ],
            },
        )

        expected_dict = {
            "content_type": "foo.ListContainer",
            "attributes": {
                "images": [
                    "isekai-blob-ref:\\file:image1.jpg",
                    "isekai-blob-ref:\\file:image2.jpg",
                ],
                "references": [
                    "isekai-resource-ref:\\gen:ref1::pk",
                    "isekai-resource-ref:\\gen:ref2::pk",
                ],
                "mixed_list": [
                    "string_value",
                    42,
                    "isekai-resource-ref:\\gen:mixed_ref::pk",
                    {"nested_ref": "isekai-blob-ref:\\url:nested.png"},
                ],
            },
        }

        assert spec.to_dict() == expected_dict

    def test_to_dict_with_tuples(self):
        spec = Spec(
            content_type="foo.TupleContainer",
            attributes={
                "tuple_refs": (
                    ResourceRef(Key(type="gen", value="tuple_ref1")).pk,
                    BlobRef(Key(type="file", value="tuple_blob.jpg")),
                    "tuple_string",
                ),
            },
        )

        expected_dict = {
            "content_type": "foo.TupleContainer",
            "attributes": {
                "tuple_refs": [
                    "isekai-resource-ref:\\gen:tuple_ref1::pk",
                    "isekai-blob-ref:\\file:tuple_blob.jpg",
                    "tuple_string",
                ],
            },
        }

        assert spec.to_dict() == expected_dict

    def test_to_dict_deeply_nested(self):
        spec = Spec(
            content_type="foo.DeepNested",
            attributes={
                "level1": {
                    "level2": {
                        "level3": {
                            "deep_ref": ResourceRef(
                                Key(type="deep", value="nested_value")
                            ).pk,
                            "deep_list": [
                                BlobRef(Key(type="nested", value="deep_blob.png")),
                                {
                                    "even_deeper": ResourceRef(
                                        Key(type="deepest", value="bottom")
                                    ).pk
                                },
                            ],
                        },
                    },
                },
            },
        )

        expected_dict = {
            "content_type": "foo.DeepNested",
            "attributes": {
                "level1": {
                    "level2": {
                        "level3": {
                            "deep_ref": "isekai-resource-ref:\\deep:nested_value::pk",
                            "deep_list": [
                                "isekai-blob-ref:\\nested:deep_blob.png",
                                {
                                    "even_deeper": "isekai-resource-ref:\\deepest:bottom::pk"
                                },
                            ],
                        },
                    },
                },
            },
        }

        assert spec.to_dict() == expected_dict

    def test_to_dict_empty_attributes(self):
        spec = Spec(
            content_type="foo.Empty",
            attributes={},
        )

        expected_dict = {
            "content_type": "foo.Empty",
            "attributes": {},
        }

        assert spec.to_dict() == expected_dict

    def test_to_dict_none_values(self):
        spec = Spec(
            content_type="foo.WithNone",
            attributes={
                "none_value": None,
                "ref_with_none": ResourceRef(Key(type="gen", value="has_none")).pk,
                "dict_with_none": {
                    "inner_none": None,
                    "inner_ref": BlobRef(Key(type="file", value="none_test.jpg")),
                },
                "list_with_none": [
                    None,
                    ResourceRef(Key(type="gen", value="in_list")).pk,
                ],
            },
        )

        expected_dict = {
            "content_type": "foo.WithNone",
            "attributes": {
                "none_value": None,
                "ref_with_none": "isekai-resource-ref:\\gen:has_none::pk",
                "dict_with_none": {
                    "inner_none": None,
                    "inner_ref": "isekai-blob-ref:\\file:none_test.jpg",
                },
                "list_with_none": [None, "isekai-resource-ref:\\gen:in_list::pk"],
            },
        }

        assert spec.to_dict() == expected_dict

    def test_from_dict_basic(self):
        data = {
            "content_type": "foo.Bar",
            "attributes": {
                "title": "Test Title",
                "image": "isekai-blob-ref:\\url:https://example.com/image.png",
                "description": "A sample description",
                "call_to_action": "isekai-resource-ref:\\gen:call_to_action_123::pk",
                "child_object": {
                    "pk": "isekai-resource-ref:\\gen:child_object_456::pk",
                    "name": "Child Object Name",
                },
            },
        }

        spec = Spec.from_dict(data)

        expected_spec = Spec(
            content_type="foo.Bar",
            attributes={
                "title": "Test Title",
                "image": BlobRef(
                    Key(type="url", value="https://example.com/image.png")
                ),
                "description": "A sample description",
                "call_to_action": ResourceRef(
                    Key(type="gen", value="call_to_action_123")
                ).pk,
                "child_object": {
                    "pk": ResourceRef(Key(type="gen", value="child_object_456")).pk,
                    "name": "Child Object Name",
                },
            },
        )

        assert spec == expected_spec

    def test_from_dict_with_lists(self):
        data = {
            "content_type": "foo.ListContainer",
            "attributes": {
                "images": [
                    "isekai-blob-ref:\\file:image1.jpg",
                    "isekai-blob-ref:\\file:image2.jpg",
                ],
                "references": [
                    "isekai-resource-ref:\\gen:ref1::pk",
                    "isekai-resource-ref:\\gen:ref2::pk",
                ],
                "mixed_list": [
                    "string_value",
                    42,
                    "isekai-resource-ref:\\gen:mixed_ref::pk",
                    {"nested_ref": "isekai-blob-ref:\\url:nested.png"},
                ],
            },
        }

        spec = Spec.from_dict(data)

        expected_spec = Spec(
            content_type="foo.ListContainer",
            attributes={
                "images": [
                    BlobRef(Key(type="file", value="image1.jpg")),
                    BlobRef(Key(type="file", value="image2.jpg")),
                ],
                "references": [
                    ResourceRef(Key(type="gen", value="ref1")).pk,
                    ResourceRef(Key(type="gen", value="ref2")).pk,
                ],
                "mixed_list": [
                    "string_value",
                    42,
                    ResourceRef(Key(type="gen", value="mixed_ref")).pk,
                    {"nested_ref": BlobRef(Key(type="url", value="nested.png"))},
                ],
            },
        )

        assert spec == expected_spec

    def test_from_dict_deeply_nested(self):
        data = {
            "content_type": "foo.DeepNested",
            "attributes": {
                "level1": {
                    "level2": {
                        "level3": {
                            "deep_ref": "isekai-resource-ref:\\deep:nested_value::pk",
                            "deep_list": [
                                "isekai-blob-ref:\\nested:deep_blob.png",
                                {
                                    "even_deeper": "isekai-resource-ref:\\deepest:bottom::pk"
                                },
                            ],
                        },
                    },
                },
            },
        }

        spec = Spec.from_dict(data)

        expected_spec = Spec(
            content_type="foo.DeepNested",
            attributes={
                "level1": {
                    "level2": {
                        "level3": {
                            "deep_ref": ResourceRef(
                                Key(type="deep", value="nested_value")
                            ).pk,
                            "deep_list": [
                                BlobRef(Key(type="nested", value="deep_blob.png")),
                                {
                                    "even_deeper": ResourceRef(
                                        Key(type="deepest", value="bottom")
                                    ).pk
                                },
                            ],
                        },
                    },
                },
            },
        )

        assert spec == expected_spec

    def test_from_dict_empty_attributes(self):
        data = {
            "content_type": "foo.Empty",
            "attributes": {},
        }

        spec = Spec.from_dict(data)

        expected_spec = Spec(
            content_type="foo.Empty",
            attributes={},
        )

        assert spec == expected_spec

    def test_from_dict_none_values(self):
        data = {
            "content_type": "foo.WithNone",
            "attributes": {
                "none_value": None,
                "ref_with_none": "isekai-resource-ref:\\gen:has_none::pk",
                "dict_with_none": {
                    "inner_none": None,
                    "inner_ref": "isekai-blob-ref:\\file:none_test.jpg",
                },
                "list_with_none": [None, "isekai-resource-ref:\\gen:in_list::pk"],
            },
        }

        spec = Spec.from_dict(data)

        expected_spec = Spec(
            content_type="foo.WithNone",
            attributes={
                "none_value": None,
                "ref_with_none": ResourceRef(Key(type="gen", value="has_none")).pk,
                "dict_with_none": {
                    "inner_none": None,
                    "inner_ref": BlobRef(Key(type="file", value="none_test.jpg")),
                },
                "list_with_none": [
                    None,
                    ResourceRef(Key(type="gen", value="in_list")).pk,
                ],
            },
        )

        assert spec == expected_spec

    def test_from_dict_invalid_ref_string(self):
        data = {
            "content_type": "foo.Invalid",
            "attributes": {
                "bad_ref": "invalid-ref-string",
            },
        }

        spec = Spec.from_dict(data)

        # Should not parse invalid ref strings, just keep them as strings
        expected_spec = Spec(
            content_type="foo.Invalid",
            attributes={
                "bad_ref": "invalid-ref-string",
            },
        )

        assert spec == expected_spec

    def test_roundtrip_to_dict_from_dict(self):
        original_spec = Spec(
            content_type="foo.Roundtrip",
            attributes={
                "title": "Roundtrip Test",
                "image": BlobRef(Key(type="url", value="https://example.com/test.jpg")),
                "refs": [
                    ResourceRef(Key(type="gen", value="ref1")).pk,
                    ResourceRef(Key(type="gen", value="ref2")).pk,
                ],
                "nested": {
                    "deep_ref": ResourceRef(Key(type="deep", value="nested")).pk,
                    "values": [
                        1,
                        2,
                        None,
                        BlobRef(Key(type="file", value="nested.png")),
                    ],
                },
            },
        )

        # Convert to dict and back
        dict_data = original_spec.to_dict()
        reconstructed_spec = Spec.from_dict(dict_data)

        assert reconstructed_spec == original_spec

    def test_find_refs(self):
        spec = Spec(
            content_type="foo.WithRefs",
            attributes={
                "title": "Test Title",
                "image": BlobRef(
                    Key(type="url", value="https://example.com/image.png")
                ),
                "call_to_action": ResourceRef(
                    Key(type="gen", value="call_to_action_123")
                ).pk,
                "child_object": {
                    "pk": ResourceRef(Key(type="gen", value="child_object_456")).pk,
                    "image": BlobRef(Key(type="file", value="child.jpg")),
                    "name": "Child Object Name",
                },
                "refs_list": [
                    ResourceRef(Key(type="gen", value="ref1")).pk,
                    BlobRef(Key(type="file", value="list_blob.jpg")),
                    "string_value",
                ],
                "duplicate_ref": ResourceRef(
                    Key(type="gen", value="call_to_action_123")
                ).pk,  # Duplicate
                "nested": {
                    "deep": {
                        "ref": ResourceRef(Key(type="gen", value="deep_ref")).pk,
                        "blob": BlobRef(Key(type="url", value="deep.png")),
                    }
                },
            },
        )

        refs = spec.find_refs()

        # Should contain all unique refs without duplicates
        expected_refs = [
            BlobRef(Key(type="url", value="https://example.com/image.png")),
            ResourceRef(Key(type="gen", value="call_to_action_123")).pk,
            ResourceRef(Key(type="gen", value="child_object_456")).pk,
            BlobRef(Key(type="file", value="child.jpg")),
            ResourceRef(Key(type="gen", value="ref1")).pk,
            BlobRef(Key(type="file", value="list_blob.jpg")),
            ResourceRef(Key(type="gen", value="deep_ref")).pk,
            BlobRef(Key(type="url", value="deep.png")),
        ]

        assert len(refs) == len(expected_refs)
        for expected_ref in expected_refs:
            assert expected_ref in refs

    def test_find_refs_no_refs(self):
        spec = Spec(
            content_type="foo.NoRefs",
            attributes={
                "title": "Test Title",
                "count": 42,
                "nested": {
                    "value": "string",
                    "list": [1, 2, "three"],
                },
            },
        )

        refs = spec.find_refs()
        assert refs == []

    def test_find_refs_with_single_string_ref(self):
        """Test finding a single ResourceRef embedded in a string."""
        author_key = Key(type="url", value="https://example.com/author/1")
        spec = Spec(
            content_type="testapp.Article",
            attributes={
                "title": f"Article by {ref(ResourceRef(author_key).name)}",
            },
        )

        refs = spec.find_refs()

        # Should find the ResourceRef embedded in the string
        expected_ref = ResourceRef(author_key).name
        assert len(refs) == 1
        assert expected_ref in refs

    def test_find_refs_with_multiple_string_refs(self):
        """Test finding multiple refs embedded in a single string."""
        author_key = Key(type="url", value="https://example.com/author/1")
        category_key = Key(type="url", value="https://example.com/category/tech")
        spec = Spec(
            content_type="testapp.Article",
            attributes={
                "description": f"Written by {ref(ResourceRef(author_key).name)} in {ref(ResourceRef(category_key).name)}",
            },
        )

        refs = spec.find_refs()

        # Should find both ResourceRefs embedded in the string
        assert len(refs) == 2
        assert ResourceRef(author_key).name in refs
        assert ResourceRef(category_key).name in refs

    def test_find_refs_with_blob_ref_in_string(self):
        """Test finding BlobRef embedded in a string."""
        image_key = Key(type="url", value="https://example.com/avatar.jpg")
        spec = Spec(
            content_type="testapp.Author",
            attributes={
                "bio": f"Profile picture: {ref(BlobRef(image_key))}",
            },
        )

        refs = spec.find_refs()

        # Should find the BlobRef embedded in the string
        expected_ref = BlobRef(image_key)
        assert len(refs) == 1
        assert expected_ref in refs

    def test_find_refs_ignores_model_ref_in_string(self):
        """Test that ModelRef in string is ignored (not a dependency)."""
        spec = Spec(
            content_type="testapp.Article",
            attributes={
                "author": ModelRef("testapp.Author", pk=42).pk,
                "description": f"By {ref(ModelRef('testapp.Author', pk=42).name)}",
            },
        )

        refs = spec.find_refs()

        # Should not find any refs (ModelRef is not a dependency)
        assert len(refs) == 0

    def test_find_refs_with_nested_string_refs(self):
        """Test finding refs in strings within nested structures."""
        author_key = Key(type="url", value="https://example.com/author/1")
        image_key = Key(type="url", value="https://example.com/image.jpg")
        spec = Spec(
            content_type="testapp.Article",
            attributes={
                "metadata": {
                    "credit": f"Photo by {ref(ResourceRef(author_key).name)}",
                    "info": {
                        "caption": f"See image at {ref(BlobRef(image_key))}",
                    },
                },
            },
        )

        refs = spec.find_refs()

        # Should find both refs in nested strings
        assert len(refs) == 2
        assert ResourceRef(author_key).name in refs
        assert BlobRef(image_key) in refs

    def test_find_refs_mixed_direct_and_string_refs(self):
        """Test finding both direct refs and refs embedded in strings."""
        author_key = Key(type="url", value="https://example.com/author/1")
        image_key = Key(type="url", value="https://example.com/image.jpg")
        category_key = Key(type="url", value="https://example.com/category/tech")

        spec = Spec(
            content_type="testapp.Article",
            attributes={
                "featured_image": BlobRef(image_key),  # Direct BlobRef
                "author": ResourceRef(author_key).pk,  # Direct ResourceRef
                "description": f"Category: {ref(ResourceRef(category_key).name)}",  # String ref
            },
        )

        refs = spec.find_refs()

        # Should find all three refs (2 direct + 1 in string)
        assert len(refs) == 3
        assert BlobRef(image_key) in refs
        assert ResourceRef(author_key).pk in refs
        assert ResourceRef(category_key).name in refs



from typing import cast
from urllib.parse import urljoin, urlparse

from bs4 import BeautifulSoup, Tag

from isekai.types import BlobResource, Key, MinedResource, TextResource


class BaseMiner:
    def mine(
        self, key: Key, resource: TextResource | BlobResource
    ) -> list[MinedResource]:
        return []


class BaseHTMLMiner(BaseMiner):
    """
    Base class for HTML miners that extract URLs with common functionality.

    Provides shared functionality for:
    - URL resolution behavior (absolute/relative)
    - Base URL determination from Host headers or URL keys
    - Domain filtering with allowlists
    - Key type determination (url vs path)

    Subclasses must implement:
    - _extract_urls(): Extract URLs and metadata from parsed HTML
    """

    allowed_domains: list[str] = []

    def __init__(self, allowed_domains: list[str] | None = None):
        """
        Initialize BaseHTMLMiner.

        Args:
            allowed_domains: Optional list of allowed domains. If empty/None,
                           all URLs are denied. Use ['*'] to allow all domains.
        """
        self.allowed_domains = allowed_domains or self.allowed_domains

    def mine(
        self, key: Key, resource: TextResource | BlobResource
    ) -> list[MinedResource]:
        mined_resources = []

        if not isinstance(resource, TextResource):
            return mined_resources

        base_url = self._determine_base_url(key, resource)
        soup = BeautifulSoup(resource.text, "html.parser")
        url_data = self._extract_urls(soup)

        for url, metadata in url_data:
            parsed_url = urlparse(url)
            if parsed_url.scheme:
                resolved_url = url
            elif base_url is None:
                resolved_url = url
            else:
                resolved_url = urljoin(base_url, url)

            if self._is_domain_allowed(resolved_url) and resolved_url:
                parsed_resolved = urlparse(resolved_url)
                if parsed_resolved.scheme:
                    mined_key = Key(type="url", value=resolved_url)
                else:
                    mined_key = Key(type="path", value=resolved_url)

                mined_resources.append(MinedResource(key=mined_key, metadata=metadata))

        return mined_resources

    def _extract_urls(self, soup: BeautifulSoup) -> list[tuple[str, dict[str, str]]]:
        """
        Extract URLs and their metadata from parsed HTML.

        Subclasses should override this method to implement their specific URL extraction logic.

        Args:
            soup: BeautifulSoup parsed HTML

        Returns:
            List of (url, metadata_dict) tuples
        """
        return []

    def _determine_base_url(
        self, key: Key, resource: TextResource | BlobResource
    ) -> str | None:
        """Determine the best base URL for resolving relative URLs."""
        if "response_headers" in resource.metadata:
            response_headers = resource.metadata["response_headers"]
            if "Host" in response_headers:
                host = response_headers["Host"]
                if key.type == "url":
                    original_url = key.value
                    parsed_original = urlparse(original_url)
                    scheme = parsed_original.scheme or "https"
                    return f"{scheme}://{host}"
                else:
                    return f"https://{host}"

        if key.type == "url":
            return key.value

        return None

    def _is_domain_allowed(self, url: str) -> bool:
        """Check if URL's domain is allowed based on allowed_domains."""
        parsed_url = urlparse(url)

        if not parsed_url.netloc:
            return True

        if not self.allowed_domains:
            return False

        if "*" in self.allowed_domains:
            return True

        return parsed_url.netloc in self.allowed_domains


class HTMLImageMiner(BaseHTMLMiner):
    """
    Extracts image URLs from HTML content with accessibility metadata.

    Parses HTML using BeautifulSoup to find image URLs from:
    - <img> src attributes
    - <img> srcset attributes
    - <source> srcset attributes (inside <picture> elements only)

    Metadata extraction:
    - Alt text from <img> tags is captured and stored in metadata["alt_text"]
    """

    def _extract_urls(self, soup: BeautifulSoup) -> list[tuple[str, dict[str, str]]]:
        """Extract image URLs and alt text from HTML."""
        image_data = []

        for img in soup.find_all("img"):
            img_tag = cast(Tag, img)
            alt_text = str(img_tag.get("alt", ""))

            src = img_tag.get("src")
            if src:
                metadata = {"alt_text": alt_text} if alt_text else {}
                image_data.append((str(src), metadata))

            srcset = img_tag.get("srcset")
            if srcset:
                metadata = {"alt_text": alt_text} if alt_text else {}
                for url in self._parse_srcset(str(srcset)):
                    image_data.append((url, metadata))

        for picture in soup.find_all("picture"):
            picture_tag = cast(Tag, picture)
            for source in picture_tag.find_all("source"):
                source_tag = cast(Tag, source)
                srcset = source_tag.get("srcset")
                if srcset:
                    for url in self._parse_srcset(str(srcset)):
                        image_data.append((url, {}))

        return image_data

    def _parse_srcset(self, srcset: str) -> list[str]:
        """Parse srcset attribute and extract URLs."""
        urls = []
        for entry in srcset.split(","):
            entry = entry.strip()
            if entry:
                url_part = entry.split()[0]
                urls.append(url_part)
        return urls


class HTMLDocumentMiner(BaseHTMLMiner):
    """
    Extracts document URLs from HTML content with link metadata.

    Parses HTML using BeautifulSoup to find document URLs from:
    - <a> href attributes pointing to document files

    Default supported document formats:
    - PDF: pdf
    - Word documents: doc, docx
    - Excel spreadsheets: xls, xlsx
    - PowerPoint presentations: ppt, pptx
    - Text files: txt
    - CSV files: csv
    - RTF files: rtf

    Metadata extraction:
    - Link text from <a> tags is captured and stored in metadata["link_text"]
    """

    document_extensions: list[str] = [
        "pdf",
        "doc",
        "docx",
        "xls",
        "xlsx",
        "ppt",
        "pptx",
        "txt",
        "csv",
        "rtf",
    ]

    def __init__(
        self,
        allowed_domains: list[str] | None = None,
        document_extensions: list[str] | None = None,
    ):
        """
        Initialize HTMLDocumentMiner.

        Args:
            allowed_domains: Optional list of allowed domains. If empty/None,
                           all URLs are denied. Use ['*'] to allow all domains.
            document_extensions: Optional list of document file extensions to look for.
                               Should be lowercase without dots (e.g., ['pdf', 'docx']).
                               If None, uses default extensions.
        """
        super().__init__(allowed_domains)
        self.document_extensions = document_extensions or self.document_extensions

    def _extract_urls(self, soup: BeautifulSoup) -> list[tuple[str, dict[str, str]]]:
        """Extract document URLs and link text from HTML."""
        document_data = []

        for link in soup.find_all("a", href=True):
            link_tag = cast(Tag, link)
            href = str(link_tag.get("href", ""))

            if not href or href.startswith(("#", "javascript:", "mailto:")):
                continue

            if self._is_document_url(href):
                link_text = link_tag.get_text(separator=" ", strip=True)
                metadata = {"link_text": link_text}
                document_data.append((href, metadata))

        return document_data

    def _is_document_url(self, url: str) -> bool:
        """Check if URL points to a document file based on file extension."""
        parsed_url = urlparse(url)
        path = parsed_url.path.lower()

        if "." in path:
            extension = path.split(".")[-1]
            return extension in self.document_extensions

        return False


class HTMLPageMiner(BaseHTMLMiner):
    """
    Extracts page URLs from HTML content for web scraping.

    Parses HTML using BeautifulSoup to find page URLs from:
    - <a> href attributes pointing to HTML pages

    Ignores links with file extensions (any link ending with .something)
    Ignores email/telephone/javascript/fragment links

    URL normalization:
    - Adds trailing slashes
    - Removes query parameters and fragments

    Dependency logic:
    - Only direct parent (immediate ancestor) is marked as dependency
    - Only applies to URL keys, non-URL keys mark all as dependencies

    Metadata extraction:
    - Link text from <a> tags is captured and stored in metadata["link_text"]
    """

    def _extract_urls(self, soup: BeautifulSoup) -> list[tuple[str, dict[str, str]]]:
        """Extract page URLs and link text from HTML."""
        page_data = []
        seen_urls = set()

        for link in soup.find_all("a", href=True):
            link_tag = cast(Tag, link)
            href = str(link_tag.get("href", ""))

            if not href or href.startswith(("#", "javascript:", "mailto:", "tel:")):
                continue

            if self._is_page_url(href):
                normalized_href = self._normalize_url(href)

                # Skip if we've already seen this URL
                if normalized_href in seen_urls:
                    continue

                seen_urls.add(normalized_href)
                link_text = link_tag.get_text(separator=" ", strip=True)
                metadata = {"link_text": link_text}
                page_data.append((normalized_href, metadata))

        return page_data

    def mine(
        self, key: Key, resource: TextResource | BlobResource
    ) -> list[MinedResource]:
        mined_resources = []

        if not isinstance(resource, TextResource):
            return mined_resources

        base_url = self._determine_base_url(key, resource)
        soup = BeautifulSoup(resource.text, "html.parser")
        url_data = self._extract_urls(soup)

        for url, metadata in url_data:
            parsed_url = urlparse(url)
            if parsed_url.scheme:
                resolved_url = url
            elif base_url is None:
                resolved_url = url
            else:
                resolved_url = urljoin(base_url, url)
                resolved_url = self._normalize_url(resolved_url)

            if self._is_domain_allowed(resolved_url) and resolved_url:
                parsed_resolved = urlparse(resolved_url)
                if parsed_resolved.scheme:
                    mined_key = Key(type="url", value=resolved_url)
                else:
                    mined_key = Key(type="path", value=resolved_url)

                mined_resources.append(MinedResource(key=mined_key, metadata=metadata))

        return mined_resources

    def _is_page_url(self, url: str) -> bool:
        """Check if URL points to a page (no file extension)."""
        parsed_url = urlparse(url)
        path = parsed_url.path

        # If path has a file extension, it's not a page
        if path and "." in path.split("/")[-1]:
            return False

        return True

    def _normalize_url(self, url: str) -> str:
        """Normalize URL by adding trailing slash and removing query/fragment."""
        parsed = urlparse(url)

        # Remove query parameters and fragments
        path = parsed.path

        # Add trailing slash if not present and path doesn't end with a file extension
        if path and not path.endswith("/") and "." not in path.split("/")[-1]:
            path += "/"
        elif not path:
            path = "/"

        # Reconstruct URL without query/fragment
        normalized = parsed._replace(path=path, query="", fragment="").geturl()
        return normalized



from django.db import models
from wagtail.models import Page

from isekai.loaders import ModelLoader
from isekai.types import Key, Resolver, ResourceRef, Spec


class PageLoader(ModelLoader):
    _parent_page_prefix = "__wagtail_parent_page"

    def load(
        self, specs: list[tuple[Key, Spec]], resolver: Resolver
    ) -> list[tuple[Key, models.Model]]:
        # Check if any of the specs are for Wagtail Page models
        if not any(spec.attributes.get(self._parent_page_prefix) for _, spec in specs):
            # If not, let another loader handle it
            return []

        page_specs = [
            (key, spec)
            for key, spec in specs
            if spec.attributes.get(self._parent_page_prefix)
        ]

        key_to_parent_page_ref = {}

        for key, spec in page_specs:
            # Remove the parent page attribute after capturing it
            # so that ModelLoader doesn't try to process it
            parent_page_ref_or_id = spec.attributes.pop(self._parent_page_prefix)
            key_to_parent_page_ref[key] = parent_page_ref_or_id

        # Let ModelLoader create the pages first
        created_objects = super().load(specs, resolver)

        # Move the created pages to their actual parent pages
        key_to_page = {
            key: obj for key, obj in created_objects if isinstance(obj, Page)
        }

        for key, _ in page_specs:
            page = key_to_page[key]
            parent_page_ref_or_id = key_to_parent_page_ref[key]

            # If the reference passed is an ID, we can fetch the page directly
            # If it's a ref, we can check if it's in this batch
            # If not, we can use resolver
            if isinstance(parent_page_ref_or_id, int):
                parent_page = Page.objects.get(pk=parent_page_ref_or_id)
            elif isinstance(parent_page_ref_or_id, ResourceRef):
                parent_ref = parent_page_ref_or_id
                if parent_ref.key in key_to_page:
                    parent_page = key_to_page[parent_ref.key]
                else:
                    # Resolve to get the parent model instance
                    resolved_parent = resolver(parent_ref)
                    # Fetch a fresh Page instance from DB to ensure tree fields are correct
                    parent_page = Page.objects.get(pk=resolved_parent.pk)
            else:
                raise ValueError(
                    f"Invalid {self._parent_page_prefix} value: {parent_page_ref_or_id}"
                )

            page.move(parent_page, pos="last-child")

        return created_objects

    def _save_object(self, model_class, obj_fields):
        """Save the object to the database."""
        # Wagtail should only have one root page
        root_page = getattr(self, "_root_page", Page.get_root_nodes().get())
        self._root_page = root_page
        assert isinstance(root_page, Page)

        obj = model_class(**obj_fields)
        root_page.add_child(instance=obj)
        return obj



from django.apps import apps

from ..models import AbstractResource


def get_resource_model() -> type[AbstractResource]:
    """
    Find the first concrete subclass of AbstractResource.

    Returns:
        Model class: The first concrete subclass of AbstractResource found.

    Raises:
        RuntimeError: If no concrete subclass of AbstractResource is found.
    """

    # Get all registered models
    for model in apps.get_models():
        # Check if the model is a subclass of AbstractResource and not abstract
        if (
            issubclass(model, AbstractResource)
            and model is not AbstractResource
            and not model._meta.abstract
        ):
            return model

    raise RuntimeError(
        "No concrete subclass of AbstractResource found. "
        "Make sure you have created a concrete model that inherits from AbstractResource."
    )



SECRET_KEY = "test-secret-key"


class DisableMigrations:
    def __contains__(self, item):
        return True

    def __getitem__(self, item):
        return None


MIGRATION_MODULES = DisableMigrations()

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
    }
}

INSTALLED_APPS = [
    "django.contrib.contenttypes",
    "django.contrib.auth",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "taggit",
    "wagtail.contrib.forms",
    "wagtail.contrib.redirects",
    "wagtail.embeds",
    "wagtail.sites",
    "wagtail.users",
    "wagtail.snippets",
    "wagtail.documents",
    "wagtail.images",
    "wagtail.search",
    "wagtail.admin",
    "wagtail",
    "isekai",
    "tests.testapp",
]

USE_TZ = True

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"

MIDDLEWARE = [
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "wagtail.contrib.redirects.middleware.RedirectMiddleware",
]

STATIC_URL = "/static/"

ROOT_URLCONF = "tests.urls"

WAGTAIL_SITE_NAME = "Test Site"



from django.contrib import admin

from isekai.models import AbstractResource


def set_status_to_extracted(modeladmin, request, queryset):
    """Set selected resources to EXTRACTED status."""
    updated = queryset.update(status=AbstractResource.Status.EXTRACTED)
    modeladmin.message_user(request, f"{updated} resources marked as EXTRACTED.")


set_status_to_extracted.short_description = "Mark as EXTRACTED"


class AbstractResourceAdmin(admin.ModelAdmin):
    list_display = [
        "key",
        "status",
        "data_type",
        "mime_type",
        "target_content_type",
        "seeded_at",
        "extracted_at",
        "mined_at",
        "transformed_at",
        "loaded_at",
    ]
    filter_horizontal = ["dependencies"]
    actions = [set_status_to_extracted]
    list_filter = [
        "status",
        "data_type",
        "mime_type",
        "target_content_type",
        "seeded_at",
    ]
    search_fields = ["key", "mime_type"]
    readonly_fields = [
        "seeded_at",
        "extracted_at",
        "mined_at",
        "transformed_at",
        "loaded_at",
    ]
    fieldsets = [
        (None, {"fields": ["key", "status", "last_error"]}),
        (
            "Data",
            {
                "fields": [
                    "data_type",
                    "mime_type",
                    "text_data",
                    "blob_data",
                    "metadata",
                ]
            },
        ),
        (
            "Dependencies",
            {
                "fields": ["dependencies"],
                "classes": ["collapse"],
            },
        ),
        (
            "Target",
            {
                "fields": ["target_content_type", "target_object_id", "target_spec"],
                "classes": ["collapse"],
            },
        ),
        (
            "Audit",
            {
                "fields": [
                    "seeded_at",
                    "extracted_at",
                    "mined_at",
                    "transformed_at",
                    "loaded_at",
                ],
                "classes": ["collapse"],
            },
        ),
    ]



import io
import os
import re
from collections.abc import Mapping
from dataclasses import dataclass
from pathlib import Path
from typing import IO, TYPE_CHECKING, Any, ClassVar, Literal, Protocol, overload
from urllib.parse import parse_qs, unquote, urlencode

if TYPE_CHECKING:
    from django.db.models import FieldFile


@dataclass(frozen=True, slots=True)
class Key:
    """
    Represents a resource key.
    """

    type: str
    value: str

    @classmethod
    def from_string(cls, key: str) -> "Key":
        """
        Parses a string into a Key object.
        """
        try:
            key, value = key.split(":", 1)

            if value == "":
                raise ValueError("Key must have a value.")

        except ValueError as err:
            raise ValueError(f"Invalid key format: {key}.") from err

        return cls(type=key, value=value)

    def __str__(self) -> str:
        """
        Returns the string representation of the Key.
        """
        return f"{self.type}:{self.value}"


@dataclass(frozen=True, slots=True)
class SeededResource:
    key: Key
    metadata: Mapping[str, Any]


@dataclass(frozen=True, slots=True)
class TextResource:
    mime_type: str
    text: str
    metadata: Mapping[str, Any]


class FileProxy(Protocol):
    @property
    def name(self) -> str: ...
    def open(self) -> IO[bytes]: ...


@dataclass(frozen=True)
class PathFileProxy:
    path: Path

    @property
    def name(self) -> str:
        return self.path.name

    def open(self) -> IO[bytes]:
        return self.path.open("rb")


@dataclass(frozen=True)
class FieldFileProxy:
    ff: "FieldFile"

    @property
    def name(self) -> str:
        return os.path.basename(self.ff.name)

    def open(self) -> IO[bytes]:
        return self.ff.storage.open(self.ff.name, mode="rb")


@dataclass(frozen=True)
class InMemoryFileProxy:
    content: bytes

    @property
    def name(self):
        return "in_memory_file"

    def open(self) -> IO[bytes]:
        return io.BytesIO(self.content)


@dataclass(frozen=True, slots=True)
class BlobResource:
    mime_type: str
    filename: str
    file_ref: FileProxy
    metadata: Mapping[str, Any]


@dataclass(frozen=True, slots=True)
class MinedResource:
    key: Key
    metadata: Mapping[str, Any]


@dataclass(frozen=True, slots=True)
class Spec:
    content_type: str
    attributes: dict[str, Any]

    def to_dict(self):
        def serialize_value(value):
            if isinstance(value, BlobRef | ResourceRef | ModelRef):
                return str(value)
            elif isinstance(value, dict):
                return {k: serialize_value(v) for k, v in value.items()}
            elif isinstance(value, list | tuple):
                return [serialize_value(v) for v in value]
            return value

        return {
            "content_type": self.content_type,
            "attributes": {
                key: serialize_value(value) for key, value in self.attributes.items()
            },
        }

    @classmethod
    def from_dict(cls, data):
        def deserialize_value(value):
            if isinstance(value, str):
                # Try to parse as various ref types
                try:
                    if value.startswith("isekai-resource-ref:\\"):
                        return ResourceRef.from_string(value)
                    elif value.startswith("isekai-blob-ref:\\"):
                        return BlobRef.from_string(value)
                    elif value.startswith("isekai-model-ref:\\"):
                        return ModelRef.from_string(value)
                except ValueError:
                    pass
                # If parsing fails or doesn't match patterns, return as string
                return value
            elif isinstance(value, dict):
                return {k: deserialize_value(v) for k, v in value.items()}
            elif isinstance(value, list | tuple):
                return [deserialize_value(v) for v in value]
            return value

        return cls(
            content_type=data["content_type"],
            attributes={
                key: deserialize_value(value)
                for key, value in data["attributes"].items()
            },
        )

    def find_refs(self) -> list["BlobRef | ResourceRef"]:
        """
        Find all Refs in the attributes dict.

        Only BlobRef and ResourceRef create resource dependencies.
        ModelRef doesn't create dependencies since it references existing DB objects directly.
        """
        refs = []
        seen = set()

        def collect_refs(value):
            if isinstance(value, BlobRef | ResourceRef):
                # Only collect BlobRef and ResourceRef, not ModelRef
                ref_str = str(value)
                if ref_str not in seen:
                    seen.add(ref_str)
                    refs.append(value)
            elif isinstance(value, str):
                # Find refs embedded in strings (created with ref() helper)
                string_refs = find_refs_in_string(value)
                for _, ref_obj in string_refs:
                    # Only collect BlobRef and ResourceRef, not ModelRef
                    if isinstance(ref_obj, BlobRef | ResourceRef):
                        ref_str = str(ref_obj)
                        if ref_str not in seen:
                            seen.add(ref_str)
                            refs.append(ref_obj)
            elif isinstance(value, dict):
                for v in value.values():
                    collect_refs(v)
            elif isinstance(value, list | tuple):
                for item in value:
                    collect_refs(item)

        collect_refs(self.attributes)
        return refs


class ModelRef:
    """
    Represents a reference to an existing database model using content_type and lookup kwargs.

    Supports lazy attribute chaining: ModelRef("app.Model", pk=42).group.name

    Will fetch the model from the database and resolve to the instance (or attribute value) during Load.
    """

    _prefix: ClassVar[str] = "isekai-model-ref:\\"

    def __init__(
        self, content_type: str, attr_path: tuple[str, ...] = (), **lookup_kwargs
    ):
        object.__setattr__(self, "_ref_content_type", content_type)
        object.__setattr__(self, "_ref_lookup_kwargs", lookup_kwargs)
        object.__setattr__(self, "_ref_attr_path", attr_path)

    @property
    def ref_content_type(self) -> str:
        return self._ref_content_type  # type: ignore

    @property
    def ref_attr_path(self) -> tuple[str, ...]:
        return self._ref_attr_path  # type: ignore

    @property
    def ref_lookup_kwargs(self) -> dict[str, Any]:
        return self._ref_lookup_kwargs  # type: ignore

    def __getattr__(self, name: str) -> "ModelRef":
        """
        Capture attribute access and return a new ModelRef with extended attr_path.
        """
        # Return new ModelRef with extended attribute path
        new_ref = ModelRef.__new__(ModelRef)
        # Access internal attributes via object.__getattribute__ to bypass __getattr__
        content_type = object.__getattribute__(self, "_ref_content_type")
        lookup_kwargs = object.__getattribute__(self, "_ref_lookup_kwargs")
        attr_path = object.__getattribute__(self, "_ref_attr_path")
        object.__setattr__(new_ref, "_ref_content_type", content_type)
        object.__setattr__(new_ref, "_ref_lookup_kwargs", lookup_kwargs)
        object.__setattr__(new_ref, "_ref_attr_path", attr_path + (name,))
        return new_ref

    def __eq__(self, other):
        if not isinstance(other, ModelRef):
            return False
        return (
            self.ref_content_type == other.ref_content_type
            and self.ref_lookup_kwargs == other.ref_lookup_kwargs
            and self.ref_attr_path == other.ref_attr_path
        )

    def __hash__(self):
        return hash(
            (
                self.ref_content_type,
                tuple(sorted(self.ref_lookup_kwargs.items())),
                self.ref_attr_path,
            )
        )

    @classmethod
    def from_string(cls, refstr: str) -> "ModelRef":
        """
        Parses a string into a ModelRef object.
        Format: "isekai-model-ref:\\app.Model?pk=42&slug=foo::attr1.attr2"
        """
        if not refstr.startswith(cls._prefix):
            raise ValueError(f"Invalid ref: {refstr}")

        # Remove prefix
        rest = refstr.removeprefix(cls._prefix)

        # Check if there's an attribute path
        if "::" in rest:
            query_part, attr_str = rest.split("::", 1)
            attr_path = tuple(attr_str.split("."))
        else:
            query_part = rest
            attr_path = ()

        # Split content_type and query string
        if "?" in query_part:
            content_type, query_string = query_part.split("?", 1)
            # Parse query string - parse_qs returns lists, we want single values
            parsed = parse_qs(query_string)
            lookup_kwargs = {k: unquote(v[0]) for k, v in parsed.items()}
        else:
            raise ValueError(
                f"Invalid ModelRef format (missing query params): {refstr}"
            )

        return cls(content_type=content_type, attr_path=attr_path, **lookup_kwargs)

    def __str__(self) -> str:
        """
        Returns the string representation of the ModelRef.
        """
        # Convert lookup_kwargs to query string
        query_string = urlencode(self.ref_lookup_kwargs)
        base = f"{self._prefix}{self.ref_content_type}?{query_string}"

        if self.ref_attr_path:
            return f"{base}::{'.'.join(self.ref_attr_path)}"
        return base


@dataclass(frozen=True, slots=True)
class BlobRef:
    """
    Represents a reference to a blob resource using a Key.

    Will be replaced by the resource's blob data during Load.
    """

    key: Key
    _prefix: ClassVar[str] = "isekai-blob-ref:\\"

    @classmethod
    def from_string(cls, refstr: str) -> "BlobRef":
        """
        Parses a string into a BlobRef object.
        """
        if not refstr.startswith(cls._prefix):
            raise ValueError(f"Invalid ref: {refstr}")

        key = Key.from_string(refstr.removeprefix(cls._prefix))
        return cls(key=key)

    def __str__(self) -> str:
        """
        Returns the string representation of the BlobRef.
        """
        return f"{self._prefix}{self.key}"


class ResourceRef:
    """
    Represents a reference to a resource using a Key with optional attribute access.

    Supports lazy attribute chaining: ResourceRef(key).group.name

    Will be replaced by the resource's model instance (or attribute value) during Load.
    """

    _prefix: ClassVar[str] = "isekai-resource-ref:\\"

    def __init__(self, key: Key, attr_path: tuple[str, ...] = ()):
        object.__setattr__(self, "_key", key)
        object.__setattr__(self, "_ref_attr_path", attr_path)

    @property
    def key(self) -> Key:
        return self._key  # type: ignore

    @property
    def ref_attr_path(self) -> tuple[str, ...]:
        return self._ref_attr_path  # type: ignore

    def __getattr__(self, name: str) -> "ResourceRef":
        """
        Capture attribute access and return a new ResourceRef with extended attr_path.
        """
        # Return new ResourceRef with extended attribute path
        new_ref = ResourceRef.__new__(ResourceRef)
        # Access internal attributes via object.__getattribute__ to bypass __getattr__
        key = object.__getattribute__(self, "_key")
        attr_path = object.__getattribute__(self, "_ref_attr_path")
        object.__setattr__(new_ref, "_key", key)
        object.__setattr__(new_ref, "_ref_attr_path", attr_path + (name,))
        return new_ref

    def __eq__(self, other):
        if not isinstance(other, ResourceRef):
            return False
        return self.key == other.key and self.ref_attr_path == other.ref_attr_path

    def __hash__(self):
        return hash((self.key, self.ref_attr_path))

    @classmethod
    def from_string(cls, refstr: str) -> "ResourceRef":
        """
        Parses a string into a ResourceRef object.
        Format: "isekai-resource-ref:\\type:value" or "isekai-resource-ref:\\type:value::attr1.attr2"
        """
        if not refstr.startswith(cls._prefix):
            raise ValueError(f"Invalid ref: {refstr}")

        # Remove prefix
        rest = refstr.removeprefix(cls._prefix)

        # Check if there's an attribute path
        if "::" in rest:
            key_str, attr_str = rest.split("::", 1)
            attr_path = tuple(attr_str.split("."))
        else:
            key_str = rest
            attr_path = ()

        key = Key.from_string(key_str)
        return cls(key=key, attr_path=attr_path)

    def __str__(self) -> str:
        """
        Returns the string representation of the ResourceRef.
        """
        base = f"{self._prefix}{self.key}"
        if self.ref_attr_path:
            return f"{base}::{'.'.join(self.ref_attr_path)}"
        return base


class Resolver(Protocol):
    """
    A resolver function that takes a ref and returns the appropriate value:
    - BlobRef -> FileProxy
    - ResourceRef -> Any (due to attribute chaining)
    - ModelRef -> Any (due to attribute chaining)
    """

    @overload
    def __call__(self, ref: BlobRef) -> FileProxy: ...
    @overload
    def __call__(self, ref: ResourceRef) -> Any: ...
    @overload
    def __call__(self, ref: ModelRef) -> Any: ...


@dataclass
class OperationResult:
    result: Literal["success", "partial_success", "failure"]
    messages: list[str]
    metadata: dict[str, Any]


class Operation(Protocol):
    def __call__(self) -> OperationResult: ...


# Exceptions
class TransitionError(Exception):
    pass


class ExtractError(Exception):
    pass


class TransformError(Exception):
    pass


def find_refs_in_string(
    text: str,
) -> list[tuple[str, "ResourceRef | ModelRef | BlobRef"]]:
    """
    Find all refs embedded in a string with %REFEND% delimiter.

    Returns list of (ref_string_with_delimiter, ref_object) tuples.

    Example:
        Input: "Hello isekai-resource-ref:\\gen:user1::name%REFEND%!"
        Output: [("isekai-resource-ref:\\gen:user1::name%REFEND%", ResourceRef(...))]
    """
    pattern = r"(isekai-(?:resource-ref|model-ref|blob-ref):\\[^%]+)%REFEND%"
    refs = []

    for match in re.finditer(pattern, text):
        ref_string = match.group(1)  # The ref without %REFEND%
        full_match = match.group(0)  # The ref with %REFEND%

        # Parse the ref string
        try:
            if ref_string.startswith("isekai-resource-ref:\\"):
                ref_obj = ResourceRef.from_string(ref_string)
                refs.append((full_match, ref_obj))
            elif ref_string.startswith("isekai-model-ref:\\"):
                ref_obj = ModelRef.from_string(ref_string)
                refs.append((full_match, ref_obj))
            elif ref_string.startswith("isekai-blob-ref:\\"):
                ref_obj = BlobRef.from_string(ref_string)
                refs.append((full_match, ref_obj))
        except ValueError:
            # Invalid ref format - skip it
            pass

    return refs


def ref(reference: "ResourceRef | ModelRef | BlobRef") -> str:
    """
    Mark end of reference for safe string interpolation.

    Use this when embedding refs in f-strings to clearly mark where the ref ends:
        f"Hello {ref(ResourceRef(key).name)}!"

    This wraps the ref with an end delimiter (%REFEND%) so the loader can
    accurately parse and replace it during the load phase.
    """
    return f"{reference}%REFEND%"



import logging
import random
import time
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Any, overload

from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.core.files.base import ContentFile
from django.db import transaction
from django.db.models import Model, Prefetch

from isekai.types import (
    BlobRef,
    BlobResource,
    FieldFileProxy,
    FileProxy,
    Key,
    MinedResource,
    ModelRef,
    OperationResult,
    PathFileProxy,
    ResourceRef,
    SeededResource,
    Spec,
    TextResource,
    TransformError,
)
from isekai.utils.core import get_resource_model
from isekai.utils.graphs import resolve_build_order

logger = logging.getLogger(__name__)


def get_created_object_stats(objects: list[Any]) -> dict[str, int]:
    """Returns a dictionary with counts of created objects by their model name."""
    stats = {}
    for obj in objects:
        model_name = obj.__class__.__name__
        stats[model_name] = stats.get(model_name, 0) + 1
    return stats


class Pipeline:
    """ETL pipeline that processes resources through seed, extract, mine, transform, load phases."""

    def __init__(
        self,
        seeders: list[Any],
        extractors: list[Any],
        miners: list[Any],
        transformers: list[Any],
        loaders: list[Any],
    ):
        self.seeders = seeders
        self.extractors = extractors
        self.miners = miners
        self.transformers = transformers
        self.loaders = loaders

    def get_configuration(self) -> dict[str, list[str]]:
        """
        Get pipeline configuration.

        Returns:
            Dictionary mapping stage names to lists of processor names.
        """
        pipeline_config = {}

        # Get seeders
        if self.seeders:
            pipeline_config["Seeders"] = [
                seeder.__class__.__name__ for seeder in self.seeders
            ]

        # Get extractors
        if self.extractors:
            pipeline_config["Extractors"] = [
                extractor.__class__.__name__ for extractor in self.extractors
            ]

        # Get miners
        if self.miners:
            pipeline_config["Miners"] = [
                miner.__class__.__name__ for miner in self.miners
            ]

        # Get transformers
        if self.transformers:
            pipeline_config["Transformers"] = [
                transformer.__class__.__name__ for transformer in self.transformers
            ]

        # Get loaders
        if self.loaders:
            pipeline_config["Loaders"] = [
                loader.__class__.__name__ for loader in self.loaders
            ]

        return pipeline_config

    def seed(self) -> OperationResult:
        """Seeds resources from various sources."""
        logger.setLevel(logging.INFO)
        Resource = get_resource_model()

        logger.info(f"Using {len(self.seeders)} seeders")

        failed_seeders = []

        seeded_resources: list[SeededResource] = []
        for seeder in self.seeders:
            try:
                resources = seeder.seed()
            except Exception:
                failed_seeders.append(seeder)
                continue

            seeded_resources.extend(seeder.seed())

        logger.info(f"Found {len(seeded_resources)} resources from seeders")

        resources = []
        for seeded_resource in seeded_resources:
            resource = Resource(
                key=str(seeded_resource.key),
                metadata=seeded_resource.metadata or None,
            )
            resources.append(resource)

            logger.info(f"Seeded resource: {seeded_resource.key}")

        logger.info("Saving seeded resources to database...")

        Resource.objects.bulk_create(resources, ignore_conflicts=True)

        logger.info(f"Seeding completed: {len(resources)} resources processed")

        messages = [
            f"Ran {len(self.seeders)} seeders",
            f"Seeded {len(resources)} resources",
        ]

        if failed_seeders:
            messages.append(f"{len(failed_seeders)} seeders failed")

        return OperationResult(
            result="success" if not failed_seeders else "partial_success",
            messages=messages,
            metadata={},
        )

    def _run_extractors(
        self, key: Key, metadata: dict[str, Any] | None, extractors: list[Any]
    ) -> TextResource | BlobResource | None:
        """Run extractors on a key and return the first successful result.

        This method is designed to run in a thread pool for parallel extraction.
        """
        for extractor in extractors:
            if extracted_resource := extractor.extract(key, metadata):
                return extracted_resource
        return None

    def extract(self) -> OperationResult:
        """Extracts data from a source."""
        logger.setLevel(logging.INFO)
        Resource = get_resource_model()

        logger.info(f"Using {len(self.extractors)} extractors")

        resources = list(Resource.objects.filter(status=Resource.Status.SEEDED))

        logger.info(f"Found {len(resources)} seeded resources to process")

        # Submit extraction tasks to thread pool
        with ThreadPoolExecutor(max_workers=10) as executor:
            # Map futures to resources
            future_to_resource = {}
            for resource in resources:
                logger.info(f"Extracting resource: {resource.key}")
                key = Key.from_string(resource.key)
                future = executor.submit(
                    self._run_extractors, key, resource.metadata, self.extractors
                )
                future_to_resource[future] = resource

            # Process results as they complete
            for future in as_completed(future_to_resource):
                resource = future_to_resource[future]

                try:
                    extracted_resource = future.result()

                    if extracted_resource:
                        resource.mime_type = extracted_resource.mime_type

                        # Merge metadata
                        if resource.metadata is None:
                            resource.metadata = {}

                        resource.metadata.update(dict(extracted_resource.metadata))

                        if isinstance(extracted_resource, TextResource):
                            resource.data_type = "text"
                            resource.text_data = extracted_resource.text

                            logger.info(
                                f"Extracted text data ({extracted_resource.mime_type}) for {resource.key}"
                            )
                        elif isinstance(extracted_resource, BlobResource):
                            resource.data_type = "blob"
                            # Read the temporary file and save it to the model's FileField
                            with extracted_resource.file_ref.open() as temp_file:
                                resource.blob_data.save(
                                    extracted_resource.filename,
                                    ContentFile(temp_file.read()),
                                    save=False,
                                )

                            # Clean up the temporary file
                            assert isinstance(
                                extracted_resource.file_ref, PathFileProxy
                            )
                            extracted_resource.file_ref.path.unlink(missing_ok=True)

                            logger.info(
                                f"Extracted blob data ({extracted_resource.mime_type}) for {resource.key}"
                            )

                    resource.transition_to(Resource.Status.EXTRACTED)
                    resource.save()

                    logger.info(f"Successfully extracted: {resource.key}")

                except Exception as e:
                    resource.last_error = f"{e.__class__.__name__}: {str(e)}"
                    resource.save()

                    logger.error(f"Failed to extract {resource.key}: {e}")

        extracted_count = sum(
            1 for r in resources if r.status == Resource.Status.EXTRACTED
        )
        error_count = sum(1 for r in resources if r.last_error)

        logger.info(
            f"Extraction completed: {extracted_count} successful, {error_count} errors"
        )

        messages = [
            f"Processed {len(resources)} resources",
            f"Extracted {extracted_count} resources",
        ]

        if error_count:
            messages.append(f"Failed to extract {error_count} resources")

        return OperationResult(
            result="success" if error_count == 0 else "partial_success",
            messages=messages,
            metadata={},
        )

    def mine(self) -> OperationResult:
        """Mines extracted resources to discover new resources."""
        logger.setLevel(logging.INFO)
        Resource = get_resource_model()

        logger.info(f"Using {len(self.miners)} miners")

        resources = Resource.objects.filter(status=Resource.Status.EXTRACTED)

        logger.info(f"Found {resources.count()} extracted resources to process")

        seeded_resource_count_before = Resource.objects.filter(
            status=Resource.Status.SEEDED
        ).count()

        for resource in resources:
            logger.info(f"Mining resource: {resource.key}")

            # Create appropriate resource object for mining
            key = Key.from_string(resource.key)
            resource_obj = resource.to_resource_dataclass()

            try:
                # Mine the resource
                mined_resources: list[MinedResource] = []

                for miner in self.miners:
                    mined_resources.extend(miner.mine(key, resource_obj))

                logger.info(
                    f"Discovered {len(mined_resources)} new resources from {resource.key}"
                )

                # Create Resource objects for new keys
                new_resources = [
                    Resource(
                        key=str(mr.key),
                        metadata=dict(mr.metadata) if mr.metadata else None,
                    )
                    for mr in mined_resources
                ]

                # Create resources that don't already exist
                Resource.objects.bulk_create(new_resources, ignore_conflicts=True)

                # Update the original resource that was mined
                resource.transition_to(Resource.Status.MINED)
                resource.save()

                # Clean up temporary file if it was a blob resource
                if isinstance(resource_obj, BlobResource) and isinstance(
                    resource_obj.file_ref, PathFileProxy
                ):
                    resource_obj.file_ref.path.unlink(missing_ok=True)

                logger.info(f"Successfully mined: {resource.key}")

            except Exception as e:
                resource.last_error = f"{e.__class__.__name__}: {str(e)}"
                resource.save()

                logger.error(f"Failed to mine {resource.key}: {e}")

        seeded_resource_count_after = Resource.objects.filter(
            status=Resource.Status.SEEDED
        ).count()

        newly_seeded_count = seeded_resource_count_after - seeded_resource_count_before
        mined_count = sum(1 for r in resources if r.status == Resource.Status.MINED)
        error_count = sum(1 for r in resources if r.last_error)

        logger.info(f"Mining completed: {mined_count} successful, {error_count} errors")

        messages = [
            f"Processed {len(resources)} resources",
            f"Mined {mined_count} resources",
        ]

        if error_count:
            messages.append(f"Failed to mine {error_count} resources")

        messages.append(f"Seeded {newly_seeded_count} new resources")

        return OperationResult(
            result="success" if not error_count else "partial_success",
            messages=messages,
            metadata={
                "newly_seeded_count": newly_seeded_count,
            },
        )

    def transform(self) -> OperationResult:
        """Transforms mined resources into target specifications."""
        logger.setLevel(logging.INFO)
        Resource = get_resource_model()

        logger.info(f"Using {len(self.transformers)} transformers")

        content_types = ContentType.objects.values_list("app_label", "model", "pk")

        ct_map = {f"{app_label}.{model}": pk for app_label, model, pk in content_types}

        resources = Resource.objects.filter(status=Resource.Status.MINED)

        logger.info(f"Found {resources.count()} mined resources to process")

        for resource in resources:
            logger.info(f"Transforming resource: {resource.key}")

            try:
                key = Key.from_string(resource.key)
                resource_obj = resource.to_resource_dataclass()

                # Use the first transformer that can handle the resource
                spec = None
                for transformer in self.transformers:
                    if spec := transformer.transform(key, resource_obj):
                        break

                if spec:
                    try:
                        content_type = ct_map[spec.content_type.lower()]
                    except KeyError as e:
                        raise TransformError(
                            f"Unknown content type: {spec.content_type}"
                        ) from e

                    spec_dict = spec.to_dict()
                    resource.target_content_type_id = content_type
                    resource.target_spec = spec_dict["attributes"]

                    # Set dependencies based on refs found in the spec
                    refs = spec.find_refs()
                    dependency_key_strings = [str(ref.key) for ref in refs]

                    # Check if all referenced resources exist
                    if dependency_key_strings:
                        existing_keys = set(
                            Resource.objects.filter(
                                key__in=dependency_key_strings
                            ).values_list("key", flat=True)
                        )
                        missing_keys = set(dependency_key_strings) - existing_keys
                        if missing_keys:
                            raise TransformError("Invalid refs found in spec")

                    resource.transition_to(Resource.Status.TRANSFORMED)

                    with transaction.atomic():
                        resource.save()
                        resource.dependencies.set(dependency_key_strings)  # type: ignore[call-arg]

                    logger.info(f"Successfully transformed: {resource.key}")
                else:
                    raise TransformError("No transformer could handle the resource")

            except Exception as e:
                resource.last_error = f"{e.__class__.__name__}: {str(e)}"
                resource.save()

                logger.error(f"Failed to transform {resource.key}: {e}")

        transformed_count = sum(
            1 for r in resources if r.status == Resource.Status.TRANSFORMED
        )
        error_count = sum(1 for r in resources if r.last_error)

        logger.info(
            f"Transform completed: {transformed_count} successful, {error_count} errors"
        )

        messages = [
            f"Processed {len(resources)} resources",
            f"Transformed {transformed_count} resources",
        ]

        if error_count:
            messages.append(f"Failed to transform {error_count} resources")

        return OperationResult(
            result="success" if error_count == 0 else "partial_success",
            messages=messages,
            metadata={},
        )

    def load(self) -> OperationResult:
        """Loads objects from resources"""
        logger.setLevel(logging.INFO)
        Resource = get_resource_model()

        logger.info(f"Using {len(self.loaders)} loaders")

        resources = Resource.objects.filter(
            status=Resource.Status.TRANSFORMED,
        ).prefetch_related(
            Prefetch(
                "dependencies",
                queryset=Resource.objects.only("key", "status"),
            )
        )

        logger.info(f"Found {resources.count()} transformed resources to process")

        # Calculate build order
        key_to_resource = {resource.key: resource for resource in resources}
        key_to_dependencies = {
            resource.key: list(resource.dependencies.all()) for resource in resources
        }
        nodes = key_to_resource.keys()
        edges = [
            (resource.key, dep.key)
            for resource in resources
            for dep in resource.dependencies.all()
            # When calculating the build order, we only need to deal with resources
            # that have not been loaded yet. Dependency resources that are already
            # loaded do not need to be taken into consideration, because we can
            # resolve to them immediately; no need for two-phase loading
            if dep.key in nodes
        ]

        graph = resolve_build_order(nodes, edges)

        # Track resources that can't be loaded due to unready dependencies.
        # Since graph is in topological order (dependencies first), cascading happens
        # automatically. when a dependency is marked unready, dependents will see it
        # in this set when processed later and also become unready. The topological
        # sort guarantees dependents are never processed before their dependencies.
        unready_resources = set()

        ready_states = (Resource.Status.LOADED, Resource.Status.TRANSFORMED)

        for node in graph:
            for resource_key in node:
                dependencies = key_to_dependencies[resource_key]
                if any(
                    dep.status not in ready_states or dep.key in unready_resources
                    for dep in dependencies
                ):
                    # If any of the resources in this node becomes unready, the
                    # whole node is unready
                    unready_resources.update(node)
                    break

        logger.info(f"Build order resolved into {len(graph)} phases")

        # Load the objects
        key_to_obj: dict[str, Model] = {}

        @overload
        def resolver(ref: BlobRef) -> FileProxy: ...
        @overload
        def resolver(ref: ResourceRef) -> Model | int | str: ...
        @overload
        def resolver(ref: ModelRef) -> Model: ...

        def resolver(
            ref: BlobRef | ResourceRef | ModelRef,
        ) -> FileProxy | Model | int | str:
            # If it's a BlobRef, we must return a FileProxy
            if type(ref) is BlobRef:
                if str(ref.key) in key_to_obj:
                    resource = key_to_resource[str(ref.key)]
                    if resource.blob_data:
                        return FieldFileProxy(ff=resource.blob_data)

                # If it's not there, then it's a reference to a resource that has
                # already been loaded
                if obj := Resource.objects.filter(key=str(ref.key)).first():
                    if obj and obj.blob_data:
                        return FieldFileProxy(ff=obj.blob_data)

            # If it's a ResourceRef, resolve to model instance or attribute
            elif type(ref) is ResourceRef:
                # Try to find the object in the pool of resources currently being loaded
                if str(ref.key) in key_to_obj:
                    obj = key_to_obj[str(ref.key)]
                    # Traverse attribute path
                    for attr in ref.ref_attr_path:
                        obj = getattr(obj, attr)
                    return obj

                # If it's not there, then it's a reference to a resource that has
                # already been loaded
                if resource := Resource.objects.filter(key=str(ref.key)).first():
                    # PK optimization: return target_object_id directly if attr_path is ("pk",)
                    if ref.ref_attr_path == ("pk",) and resource.target_object_id:
                        return resource.target_object_id
                    # Otherwise fetch target_object and traverse
                    if obj := resource.target_object:
                        for attr in ref.ref_attr_path:
                            obj = getattr(obj, attr)
                        return obj

            # If it's a ModelRef, fetch from DB using content_type and lookup_kwargs
            elif type(ref) is ModelRef:
                app_label, model_name = ref.ref_content_type.split(".", 1)
                model_class = apps.get_model(app_label, model_name)
                obj = model_class.objects.get(**ref.ref_lookup_kwargs)
                # Traverse attribute path
                for attr in ref.ref_attr_path:
                    obj = getattr(obj, attr)
                return obj

            # If the framework is working correctly, it is logically impossible to
            # reach this case. The build order resolver ensures that all dependency
            # resources are created either before it's needed, or at the same time
            # as the resource that references it.
            raise ValueError(f"Unable to resolve reference: {ref}")

        for node in graph:
            # Each node in the graph is comprised of one OR MORE resources.

            # If any resource in this node is unready, skip the entire node
            if any(rkey in unready_resources for rkey in node):
                logger.warning(f"Skipping node with unready dependencies: {list(node)}")
                continue

            # If there is more than one resource, that means that those resources
            # need to be loaded together using a two-phase loading process in order
            # to resolve circular dependencies.
            # Single resources are also loaded, but they don't need the same
            # circular dependency handling.
            if len(node) == 1:
                logger.info(f"Loading resource: {list(node)[0]}")
            else:
                logger.info(
                    f"Loading {len(node)} resources with circular dependencies: {list(node)}"
                )

            specs = []
            for resource_key in node:
                resource = key_to_resource[resource_key]
                ct = resource.target_content_type
                assert ct, "Resource must have a target content type to be loaded"
                model_class = ct.model_class()
                assert model_class, "Unable to resolve model class for content type"

                key = Key.from_string(resource.key)
                spec = Spec.from_dict(
                    {
                        "content_type": f"{model_class._meta.label}",
                        "attributes": resource.target_spec,
                    }
                )
                specs.append((key, spec))

            # Load the objects
            try:
                with transaction.atomic():
                    resources_to_update = []
                    created_objects = []
                    for loader in self.loaders:
                        if created_objects := loader.load(specs, resolver):
                            break

                    for ckey, cobject in created_objects:
                        key_to_obj[str(ckey)] = cobject
                        resource = key_to_resource[str(ckey)]
                        resource.target_object_id = cobject.pk
                        resource.transition_to(Resource.Status.LOADED)
                        resources_to_update.append(resource)

                        logger.info(f"Successfully loaded: {resource.key}")

                    Resource.objects.bulk_update(
                        resources_to_update,
                        [
                            "target_object_id",
                            "status",
                            "loaded_at",
                            "last_error",
                        ],
                    )
            except Exception as e:
                # Mark resources in this node as failed
                failed_resources = []
                for resource_key in node:
                    resource = key_to_resource[resource_key]
                    resource.refresh_from_db()
                    resource.last_error = f"{e.__class__.__name__}: {str(e)}"
                    failed_resources.append(resource)

                    logger.error(f"Failed to load {resource.key}: {e}")

                # Save the failed resources
                Resource.objects.bulk_update(
                    failed_resources,
                    ["last_error"],
                )

                # Stop processing - dependent nodes will also fail
                logger.error(
                    "Stopping load process due to node failure - remaining nodes would likely fail due to missing dependencies"
                )

                return OperationResult(
                    result="failure",
                    messages=[
                        f"Load failed at node with resources: {list(node)}",
                        f"Error: {e.__class__.__name__}: {str(e)}",
                    ],
                    metadata={
                        "object_stats": get_created_object_stats(
                            list(key_to_obj.values())
                        )
                    },
                )

        all_resources = list(key_to_resource.values())
        loaded_count = sum(
            1 for r in all_resources if r.status == Resource.Status.LOADED
        )
        skipped_count = sum(1 for r in all_resources if r.key in unready_resources)

        logger.info(
            f"Load completed: {loaded_count} successful, {skipped_count} skipped"
        )

        messages = [
            f"Processed {len(all_resources)} resources",
            f"Loaded {loaded_count} resources",
        ]

        if skipped_count > 0:
            messages.append(
                f"Skipped {skipped_count} resources due to unready dependencies"
            )

        return OperationResult(
            result="success",
            messages=messages,
            metadata={
                "object_stats": get_created_object_stats(list(key_to_obj.values()))
            },
        )


class DummyPipeline(Pipeline):
    """A dummy pipeline that simulates operations without actually doing any work."""

    def __init__(self, simulate_errors: bool = False, simulate_warnings: bool = False):
        # Create dummy processors
        dummy_processors = [
            type("DummyProcessor", (), {"__name__": f"Processor{i}"})()
            for i in range(2)
        ]

        super().__init__(
            seeders=dummy_processors,
            extractors=dummy_processors,
            miners=dummy_processors,
            transformers=dummy_processors,
            loaders=dummy_processors,
        )
        self.simulate_errors = simulate_errors
        self.simulate_warnings = simulate_warnings

    def get_configuration(self) -> dict[str, list[str]]:
        """Return dummy configuration."""
        return {
            "Seeders": ["DummySeeder", "AnotherDummySeeder"],
            "Extractors": ["DummyExtractor", "FakeExtractor"],
            "Miners": ["DummyMiner"],
            "Transformers": ["DummyTransformer", "FakeTransformer"],
            "Loaders": ["DummyLoader"],
        }

    def _simulate_work(
        self, operation_name: str, count: int | None = None
    ) -> tuple[int, int, int]:
        """Simulate work and return (processed, successful, errors)."""
        if count is None:
            count = random.randint(5, 20)

        # Simulate some time passing
        time.sleep(random.uniform(0.1, 0.5))

        # Map operation names to attribute names
        attr_map = {
            "seed": "seeders",
            "extract": "extractors",
            "mine": "miners",
            "transform": "transformers",
            "load": "loaders",
        }
        attr_name = attr_map.get(operation_name.lower(), operation_name.lower() + "s")
        processors = getattr(self, attr_name, [])
        logger.info(f"Using {len(processors)} {operation_name.lower()}s")
        logger.info(f"Found {count} resources to process")

        # Simulate processing each resource
        successful = 0
        errors = 0

        for i in range(count):
            # Simulate some processing time
            time.sleep(random.uniform(0.01, 0.05))

            # Simulate success/failure based on flags
            if self.simulate_errors and random.random() < 0.2:  # 20% error rate
                logger.error(f"Failed to process resource {i}: Simulated error")
                errors += 1
            else:
                logger.info(f"Successfully processed resource {i}")
                successful += 1

        logger.info("Saving changes to database...")
        logger.info(
            f"{operation_name} completed: {successful} successful, {errors} errors"
        )

        return count, successful, errors

    def seed(self) -> OperationResult:
        """Simulate seeding operation."""
        logger.setLevel(logging.INFO)

        _processed, successful, errors = self._simulate_work("Seed")

        messages = [
            f"Ran {len(self.seeders)} seeders",
            f"Seeded {successful} resources",
        ]

        if errors:
            messages.append(f"{errors} seeders failed")

        if self.simulate_errors and errors > 0:
            result = "partial_success"
        else:
            result = "success"

        return OperationResult(
            result=result,
            messages=messages,
            metadata={"seeded_count": successful},
        )

    def extract(self) -> OperationResult:
        """Simulate extraction operation."""
        logger.setLevel(logging.INFO)

        processed, successful, errors = self._simulate_work("Extract")

        messages = [
            f"Processed {processed} resources",
            f"Extracted {successful} resources",
        ]

        if errors:
            messages.append(f"Failed to extract {errors} resources")

        result = "success" if errors == 0 else "partial_success"

        return OperationResult(
            result=result,
            messages=messages,
            metadata={},
        )

    def mine(self) -> OperationResult:
        """Simulate mining operation."""
        logger.setLevel(logging.INFO)

        processed, successful, errors = self._simulate_work("Mine")
        # 50% chance of seeding 0 new resources to prevent infinite loop
        newly_seeded = 0 if random.random() < 0.5 else random.randint(2, 8)

        messages = [
            f"Processed {processed} resources",
            f"Mined {successful} resources",
        ]

        if errors:
            messages.append(f"Failed to mine {errors} resources")

        messages.append(f"Seeded {newly_seeded} new resources")

        result = "success" if errors == 0 else "partial_success"

        return OperationResult(
            result=result,
            messages=messages,
            metadata={"newly_seeded_count": newly_seeded},
        )

    def transform(self) -> OperationResult:
        """Simulate transformation operation."""
        logger.setLevel(logging.INFO)

        processed, successful, errors = self._simulate_work("Transform")

        messages = [
            f"Processed {processed} resources",
            f"Transformed {successful} resources",
        ]

        if errors:
            messages.append(f"Failed to transform {errors} resources")

        result = "success" if errors == 0 else "partial_success"

        return OperationResult(
            result=result,
            messages=messages,
            metadata={},
        )

    def load(self) -> OperationResult:
        """Simulate loading operation."""
        logger.setLevel(logging.INFO)

        processed, successful, _errors = self._simulate_work("Load")

        # Simulate object creation stats
        object_stats = {
            "Article": random.randint(3, 8),
            "Image": random.randint(1, 5),
            "Author": random.randint(1, 3),
        }

        logger.info(f"Build order resolved into {random.randint(1, 3)} phases")

        for i in range(processed):
            if i < successful:
                logger.info(f"Loading resource: dummy_resource_{i}")
                logger.info(f"Successfully loaded: dummy_resource_{i}")

        messages = [
            f"Processed {processed} resources",
            f"Loaded {successful} resources",
        ]

        if self.simulate_errors and random.random() < 0.1:  # 10% chance of failure
            return OperationResult(
                result="failure",
                messages=[
                    "Load failed at node with resources: ['dummy_resource_1']",
                    "Error: SimulatedError: Database connection failed",
                ],
                metadata={"object_stats": object_stats},
            )

        result = "success"

        return OperationResult(
            result=result,
            messages=messages,
            metadata={"object_stats": object_stats},
        )


def get_dummy_pipeline(
    simulate_errors: bool = False, simulate_warnings: bool = False
) -> DummyPipeline:
    """Returns a DummyPipeline instance for testing."""
    return DummyPipeline(
        simulate_errors=simulate_errors, simulate_warnings=simulate_warnings
    )


def get_django_pipeline() -> Pipeline:
    """Returns a Pipeline instance configured using the Django Resource model"""
    Resource = get_resource_model()

    return Pipeline(
        seeders=Resource.seeders,
        extractors=Resource.extractors,
        miners=Resource.miners,
        transformers=Resource.transformers,
        loaders=Resource.loaders,
    )
